# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
import warnings

import jax
from jax import numpy as jnp
from flax import struct

import netket.jax as nkjax
from netket.utils import timing, HashablePartial
from netket.utils.types import PyTree
from netket.utils.api_utils import partial_from_kwargs
from netket.nn import split_array_mpi
from netket.errors import (
    IllegalHolomorphicDeclarationForRealParametersError,
    NonHolomorphicQGTOnTheFlyDenseRepresentationError,
    HolomorphicUndeclaredWarning,
)

from .common import check_valid_vector_type
from .qgt_onthefly_logic import mat_vec_factory, mat_vec_chunked_factory

from ..linear_operator import LinearOperator, Uninitialized


@partial_from_kwargs
def QGTOnTheFly(
    vstate, *, chunk_size=None, holomorphic: bool | None = None, **kwargs
) -> "QGTOnTheFlyT":
    """
    Lazy representation of an S Matrix computed by performing 2 jvp
    and 1 vjp products, using the variational state's model, the
    samples that have already been computed, and the vector.

    The S matrix is not computed yet, but can be computed by calling
    :code:`to_dense`.
    The details on how the ⟨S⟩⁻¹⟨F⟩ system is solved are contained in
    the field `sr`.

    Args:
        vstate: The variational State.
        chunk_size: If supplied, overrides the chunk size of the variational state
                    (useful for models where the backward pass requires more
                    memory than the forward pass).
    """
    if kwargs.pop("diag_scale", None) is not None:
        raise NotImplementedError(
            "\n`diag_scale` argument is not yet supported by QGTOnTheFly."
            "Please use `QGTJacobianPyTree` or `QGTJacobianDense`.\n\n"
            "You are also encouraged to nag the developers to support "
            "this feature.\n\n"
        )

    # TODO: Find a better way to handle this case
    from netket.vqs import FullSumState

    if isinstance(vstate, FullSumState):
        samples = split_array_mpi(vstate._all_states)
        pdf = split_array_mpi(vstate.probability_distribution())
    else:
        samples = vstate.samples
        pdf = None

    if chunk_size is None:
        chunk_size = getattr(vstate, "chunk_size", None)

    return QGTOnTheFly_DefaultConstructor(
        vstate._apply_fun,
        vstate.parameters,
        vstate.model_state,
        samples,
        pdf=pdf,
        chunk_size=chunk_size,
        holomorphic=holomorphic,
        **kwargs,
    )


@timing.timed
def QGTOnTheFly_DefaultConstructor(
    apply_fun,
    parameters,
    model_state,
    samples,
    pdf=None,
    *,
    chunk_size: int | None = None,
    holomorphic: bool | None = None,
    **kwargs,
) -> "QGTOnTheFlyT":
    """ """
    if pdf is not None:
        if not pdf.shape == samples.shape[:-1]:
            raise ValueError(
                "The shape of pdf must match the shape of the samples, "
                f"instead you provided (pdf.shape={pdf.shape}) != "
                f"(samples.shape={samples.shape[:-1]})"
            )
        if pdf.ndim >= 2:
            pdf = jax.jit(jax.lax.collapse, static_argnums=(1, 2))(pdf, 0, 2)

    # The code does not support an extra batch dimension
    if samples.ndim >= 3:
        # use jit so that we can do it on global shared array
        samples = jax.jit(jax.lax.collapse, static_argnums=(1, 2))(samples, 0, 2)

    n_samples_per_rank = samples.shape[0] // nkjax.sharding.device_count_per_rank()
    if chunk_size is None or chunk_size >= n_samples_per_rank:
        mv_factory = mat_vec_factory
        chunking = False
    else:
        mv_factory = HashablePartial(mat_vec_chunked_factory, chunk_size=chunk_size)
        chunking = True

    # check if holomorphic or not
    if holomorphic:
        if nkjax.tree_leaf_isreal(parameters):
            raise IllegalHolomorphicDeclarationForRealParametersError()
        else:
            mode = "holomorphic"
    else:
        if not nkjax.tree_leaf_iscomplex(parameters):
            mode = "real"
        else:
            if holomorphic is None:
                warnings.warn(HolomorphicUndeclaredWarning(), UserWarning)
            mode = "complex"

    nkjax.jacobian_default_mode(
        apply_fun,
        parameters,
        model_state,
        samples,
        holomorphic=holomorphic,
    )

    mat_vec = mv_factory(
        forward_fn=apply_fun,
        params=parameters,
        model_state=model_state,
        samples=samples,
        pdf=pdf,
    )
    return QGTOnTheFlyT(
        _mat_vec=mat_vec,
        _params=parameters,
        _chunking=chunking,
        _mode=mode,
        **kwargs,
    )


@struct.dataclass
class QGTOnTheFlyT(LinearOperator):
    """
    Lazy representation of an S Matrix computed by performing 2 jvp
    and 1 vjp products, using the variational state's model, the
    samples that have already been computed, and the vector.

    The S matrix is not computed yet, but can be computed by calling
    :code:`to_dense`.
    The details on how the ⟨S⟩⁻¹⟨F⟩ system is solved are contained in
    the field `sr`.
    """

    _mat_vec: Callable[[PyTree, float], PyTree] = Uninitialized
    """The S matrix-vector product as generated by mat_vec_factory.
    It's a jax.Partial, so can be used as pytree_node."""

    _params: PyTree = Uninitialized
    """The first input to apply_fun (parameters of the ansatz).
    Only used as a shape placeholder."""

    _chunking: bool = struct.field(pytree_node=False, default=False)
    """Whether the implementation with chunks is used which currently does not support vmapping over it"""

    _mode: str = struct.field(pytree_node=False, default=None)
    """Differentiation mode:
        - "real": for real-valued R->R and C->R Ansätze, splits the complex inputs
                  into real and imaginary part.
        - "complex": for complex-valued R->C and C->C Ansätze, splits the complex
                  inputs and outputs into real and imaginary part
        - "holomorphic": for any Ansätze. Does not split complex values.
        - "auto": autoselect real or complex.
    """

    def __matmul__(self, y):
        return onthefly_mat_treevec(self, y)

    def _solve(self, solve_fun, y: PyTree, *, x0: PyTree | None, **kwargs) -> PyTree:
        return _solve(self, solve_fun, y, x0=x0)

    def to_dense(self) -> jnp.ndarray:
        """
        Convert the lazy matrix representation to a dense matrix representation.

        Returns:
            A dense matrix representation of this S matrix.
        """
        # This condition will be true if the user specified `holomorphic=False` and
        # if the parameters are complex. If the parameters are real and the user
        # did not specify holomorphic we will have mode==real and if holomorphic is
        # True mode==holomorphic.
        #
        # We must check this because the AD implementation will compute the wrong
        # QGT in that case
        if self._mode == "complex":
            raise NonHolomorphicQGTOnTheFlyDenseRepresentationError()

        return _to_dense(self)

    def __repr__(self):
        return f"QGTOnTheFly(diag_shift={self.diag_shift})"


@jax.jit
def onthefly_mat_treevec(
    S: QGTOnTheFly, vec: PyTree | jnp.ndarray
) -> PyTree | jnp.ndarray:
    """
    Perform the lazy mat-vec product, where vec is either a tree with the same structure as
    params or a ravelled vector
    """

    # if has a ndim it's an array and not a pytree
    if hasattr(vec, "ndim"):
        if not vec.ndim == 1:
            raise ValueError("Unsupported mat-vec for chunks of vectors")
        # If the input is a vector
        if not nkjax.tree_size(S._params) == vec.size:
            raise ValueError(
                """Size mismatch between number of parameters ({nkjax.tree_size(S.params)})
                                and vector size {vec.size}.
                             """
            )

        _, unravel = nkjax.tree_ravel(S._params)
        vec = unravel(vec)
        ravel_result = True
    else:
        ravel_result = False

    check_valid_vector_type(S._params, vec)

    vec = nkjax.tree_cast(vec, S._params)

    res = S._mat_vec(vec, S.diag_shift)

    if ravel_result:
        res, _ = nkjax.tree_ravel(res)

    return res


@jax.jit
def _solve(
    self: QGTOnTheFlyT, solve_fun, y: PyTree, *, x0: PyTree | None, **kwargs
) -> PyTree:
    check_valid_vector_type(self._params, y)

    y = nkjax.tree_cast(y, self._params)

    # we could cache this...
    if x0 is None:
        x0 = jax.tree_util.tree_map(jnp.zeros_like, y)

    out, info = solve_fun(self, y, x0=x0)
    return out, info


@jax.jit
def _to_dense(self: QGTOnTheFlyT) -> jnp.ndarray:
    """
    Convert the lazy matrix representation to a dense matrix representation

    Returns:
        A dense matrix representation of this S matrix.
    """
    Npars = nkjax.tree_size(self._params)
    I = jax.numpy.eye(Npars)

    if self._chunking:
        # the linear_call in mat_vec_chunked does currently not have a jax batching rule,
        # so it cannot be vmapped but we can use scan
        # which is better for reducing the memory consumption anyway
        _, out = jax.lax.scan(lambda _, x: (None, self @ x), None, I)
    else:
        out = jax.vmap(lambda x: self @ x, in_axes=0)(I)

    if jnp.iscomplexobj(out):
        out = out.T

    return out



# Copyright 2020, 2021  The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable

from functools import partial

import jax
import jax.numpy as jnp

from netket import stats
from netket.operator import AbstractOperator
from netket.optimizer.qgt import QGTJacobianDense
from netket.optimizer.qgt.qgt_jacobian_dense import convert_tree_to_dense_format
from netket.vqs import VariationalState, VariationalMixedState, MCState
from netket.jax import tree_cast
from netket.utils import timing
from netket.experimental.dynamics import AbstractSolver


from .tdvp_common import TDVPBaseDriver, odefun


class TDVPSchmitt(TDVPBaseDriver):
    r"""
    Variational time evolution based on the time-dependent variational principle which,
    when used with Monte Carlo sampling via :class:`netket.vqs.MCState`, is the time-dependent VMC
    (t-VMC) method.

    This driver, which only works with standard MCState variational states, uses the regularization
    procedure described in `M. Schmitt's PRL 125 <https://journals.aps.org/prl/pdf/10.1103/PhysRevLett.125.100503>`_ .

    With the force vector

    .. math::

        F_k=\langle \mathcal O_{\theta_k}^* E_{loc}^{\theta}\rangle_c

    and the quantum Fisher matrix

    .. math::

        S_{k,k'} = \langle \mathcal O_{\theta_k} (\mathcal O_{\theta_{k'}})^*\rangle_c

    and for real parameters :math:`\theta\in\mathbb R`, the TDVP equation reads

    .. math::

        q\big[S_{k,k'}\big]\theta_{k'} = -q\big[xF_k\big]

    Here, either :math:`q=\text{Re}` or :math:`q=\text{Im}` and :math:`x=1` for ground state
    search or :math:`x=i` (the imaginary unit) for real time dynamics.

    For ground state search a regularization controlled by a parameter :math:`\rho` can be included
    by increasing the diagonal entries and solving

    .. math::

        q\big[(1+\rho\delta_{k,k'})S_{k,k'}\big]\theta_{k'} = -q\big[F_k\big]

    The `TDVP` class solves the TDVP equation by computing a pseudo-inverse of :math:`S` via
    eigendecomposition yielding

    .. math::

        S = V\Sigma V^\dagger

    with a diagonal matrix :math:`\Sigma_{kk}=\sigma_k`
    Assuming that :math:`\sigma_1` is the smallest eigenvalue, the pseudo-inverse is constructed
    from the regularized inverted eigenvalues

    .. math::

        \tilde\sigma_k^{-1}=\frac{1}{\Big(1+\big(\frac{\epsilon_{SVD}}{\sigma_j/\sigma_1}\big)^6\Big)\Big(1+\big(\frac{\epsilon_{SNR}}{\text{SNR}(\rho_k)}\big)^6\Big)}

    with :math:`\text{SNR}(\rho_k)` the signal-to-noise ratio of
    :math:`\rho_k=V_{k,k'}^{\dagger}F_{k'}` (see
    `[arXiv:1912.08828] <https://arxiv.org/pdf/1912.08828.pdf>`_ for details).


    .. note::

        This TDVP Driver uses the time-integrators from the `nkx.dynamics` module, which are
        automatically executed under a `jax.jit` context.

        When running computations on GPU, this can lead to infinite hangs or extremely long
        compilation times. In those cases, you might try setting the configuration variable
        `nk.config.netket_experimental_disable_ode_jit = True` to mitigate those issues.

    """

    def __init__(
        self,
        operator: AbstractOperator,
        variational_state: VariationalState,
        # TODO: remove default None once `integrator` is removed
        ode_solver: AbstractSolver = None,
        *,
        t0: float = 0.0,
        propagation_type: str = "real",
        # TODO: integrator deprecated in 3.16 (oct/nov 2024)
        integrator: AbstractSolver = None,
        holomorphic: bool | None = None,
        diag_shift: float = 0.0,
        diag_scale: float | None = None,
        error_norm: str | Callable = "qgt",
        rcond: float = 1e-14,
        rcond_smooth: float = 1e-8,
        snr_atol: float = 1,
    ):
        r"""
        Initializes the time evolution driver.

        Args:
            operator: The generator of the dynamics (Hamiltonian for pure states,
                Lindbladian for density operators).
            variational_state: The variational state.
            ode_solver: Solving algorithm used the ODE.
            t0: Initial time at the start of the time evolution.
            propagation_type: Determines the equation of motion: "real"  for the
                real-time Schrödinger equation (SE), "imag" for the imaginary-time SE.
            error_norm: Norm function used to calculate the error with adaptive integrators.
                Can be either "euclidean" for the standard L2 vector norm :math:`w^\dagger w`,
                "maximum" for the maximum norm :math:`\max_i |w_i|`
                or "qgt", in which case the scalar product induced by the QGT :math:`S` is used
                to compute the norm :math:`\Vert w \Vert^2_S = w^\dagger S w` as suggested
                in PRL 125, 100503 (2020).
                Additionally, it possible to pass a custom function with signature
                :code:`norm(x: PyTree) -> float`
                which maps a PyTree of parameters :code:`x` to the corresponding norm.
                Note that norm is used in jax.jit-compiled code.
            holomorphic: a flag to indicate that the wavefunction is holomorphic.
            diag_shift: diagonal shift of the quantum geometric tensor (QGT)
            diag_scale: If not None rescales the diagonal shift of the QGT
            rcond : Cut-off ratio for small singular :math:`\sigma_k` values of the
                Quantum Geometric Tensor. For the purposes of rank determination,
                singular values are treated as zero if they are smaller than rcond times
                the largest singular value :code:`\sigma_{max}`.
            rcond_smooth : Smooth cut-off ratio for singular values of the Quantum Geometric
                Tensor. This regularization parameter used with a similar effect to `rcond`
                but with a softer curve. See :math:`\epsilon_{SVD}` in the formula
                above.
            snr_atol: Noise regularisation absolute tolerance, meaning that eigenvalues of
                the S matrix that have a signal to noise ratio above this quantity will be
                (soft) truncated. This is :math:`\epsilon_{SNR}` in the formulas above.

        """
        self.propagation_type = propagation_type
        if isinstance(variational_state, VariationalMixedState):
            # assuming Lindblad Dynamics
            # TODO: support density-matrix imaginary time evolution
            if propagation_type == "real":
                self._loss_grad_factor = 1.0
            else:
                raise ValueError(
                    "only real-time Lindblad evolution is supported for " "mixed states"
                )
        else:
            if propagation_type == "real":
                self._loss_grad_factor = -1.0j
            elif propagation_type == "imag":
                self._loss_grad_factor = -1.0
            else:
                raise ValueError("propagation_type must be one of 'real', 'imag'")

        self.rcond = rcond
        self.rcond_smooth = rcond_smooth
        self.snr_atol = snr_atol

        self.diag_shift = diag_shift
        self.holomorphic = holomorphic
        self.diag_scale = diag_scale

        super().__init__(
            operator,
            variational_state,
            ode_solver,
            t0=t0,
            error_norm=error_norm,
            integrator=integrator,
        )


# Copyright notice:
# The function `_impl` below includes lines copied from the jVMC repository
# found at github.com/markusschmitt/vmc_jax and licensed according to
# MIT License, Copyright (c) 2021 Markus Schmitt


@timing.timed
@partial(jax.jit, static_argnames=("n_samples"))
def _impl(parameters, n_samples, E_loc, S, rhs_coeff, rcond, rcond_smooth, snr_atol):
    E = stats.statistics(E_loc)
    ΔE_loc = E_loc.reshape(-1, 1) - E.mean

    stack_jacobian = S.mode == "complex"

    O = S.O / jnp.sqrt(n_samples)  # already divided by jnp.sqrt(n_s)
    if stack_jacobian:
        O = O.reshape(-1, 2, S.O.shape[-1])
        O = O[:, 0, :] + 1j * O[:, 1, :]

    Sd = S.to_dense()
    ev, V = jnp.linalg.eigh(Sd)

    OEdata = O.conj() * ΔE_loc
    F = stats.sum(OEdata, axis=0)

    # Note: this implementation differs from Eq. 20 in Markus's paper, which I would
    # implement as `rho = mpi.mean(QEdata, axis=0)`. However, this is different from
    # changing the basis AFTER averaging over the samples, and leads to the wrong
    # normalisation of RHo.
    Q = jnp.tensordot(V.conj().T, O.T, axes=1).T
    QEdata = Q.conj() * ΔE_loc
    rho = V.conj().T @ F

    # Compute the SNR according to Eq. 21
    snr = jnp.abs(rho) * jnp.sqrt(n_samples) / jnp.sqrt(stats.var(QEdata, axis=0))

    # Discard eigenvalues below numerical precision
    ev_inv = jnp.where(jnp.abs(ev / ev[-1]) > rcond, 1.0 / ev, 0.0)
    # Set regularizer for singular value cutoff
    regularizer = 1.0 / (1.0 + (rcond_smooth / jnp.abs(ev / ev[-1])) ** 6)
    # Construct a soft cutoff based on the SNR
    regularizer2 = regularizer * (1.0 / (1.0 + (snr_atol / snr) ** 6))

    # solve the linear system by hand
    eta_p = ev_inv * regularizer2 * rhs_coeff * rho
    # convert back to the parameter space
    update = V @ eta_p

    # remainder of the solution
    rmd = jnp.linalg.norm(Sd.dot(update) - rhs_coeff * F) / jnp.linalg.norm(F)

    y, reassemble = convert_tree_to_dense_format(parameters, S.mode)
    update_tree = reassemble(update if jnp.iscomplexobj(y) else update.real)

    # If parameters are real, then take only real part of the gradient (if it's complex)
    dw = tree_cast(update_tree, parameters)

    return E, dw, rmd, snr


@odefun.dispatch
def odefun_schmitt(state: MCState, self: TDVPSchmitt, t, w, *, stage=0):  # noqa: F811
    # pylint: disable=protected-access

    state.parameters = w
    state.reset()

    op_t = self.generator(t)

    E_loc = state.local_estimators(op_t)

    self._S = QGTJacobianDense(
        state,
        diag_shift=self.diag_shift,
        diag_scale=self.diag_scale,
        holomorphic=self.holomorphic,
    )

    self._loss_stats, self._dw, self._rmd, self._snr = _impl(
        state.parameters,
        state.n_samples,
        E_loc,
        self._S,
        self._loss_grad_factor,
        self.rcond,
        self.rcond_smooth,
        self.snr_atol,
    )

    if stage == 0:  # TODO: This does not work with FSAL.
        self._last_qgt = self._S

    return self._dw


@partial(jax.jit, static_argnums=(3, 4))
def _map_parameters(forces, parameters, loss_grad_factor, propagation_type, state_T):
    forces = jax.tree_util.tree_map(
        lambda x, target: loss_grad_factor * x,
        forces,
        parameters,
    )

    forces = tree_cast(forces, parameters)

    return forces



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from typing import Union
from collections.abc import Iterator, Sequence


Edge = tuple[int, int]
ColoredEdge = tuple[int, int, int]
EdgeSequence = Union[Sequence[Edge], Sequence[ColoredEdge]]


class AbstractGraph(abc.ABC):
    """Abstract class for NetKet graph objects."""

    @abc.abstractmethod
    def is_connected(self) -> bool:
        r"""True if the graph is connected"""
        raise NotImplementedError

    @abc.abstractmethod
    def is_bipartite(self) -> bool:
        r"""True if the graph is bipartite"""
        raise NotImplementedError

    @abc.abstractmethod
    def edges(
        self, return_color: bool = False, filter_color: int | None = None
    ) -> EdgeSequence:
        r"""Returns the sequence of edges of the graph.

        Arguments:
            return_color: If :code:`True`, return edges with added color information.
            filter_color: If set, return only edges of the given color.

        Returns:
            A sequence of edges as tuples :code:`(i, j)` or, if `return_color` is
            passed, a sequence of tuples :code:`(i, j, c)` with :code:`c` indicating
            the color of the respective edge.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def edge_colors(self) -> Sequence[int]:
        r"""Sequence of edge colors, in the order of the edges returned by
        :code:`self.edges`."""
        raise NotImplementedError

    @abc.abstractmethod
    def nodes(self) -> Iterator[int]:
        r"""Iterator over the nodes of the graph"""
        raise NotImplementedError

    @abc.abstractmethod
    def distances(self) -> list[list]:
        r"""List containing the distances between the nodes.
        The distance between unconnected nodes (no path exists between them) is set to -1
        """
        raise NotImplementedError

    @abc.abstractmethod
    def automorphisms(self):
        r"""Symmetry group containing the automorphisms of the graph"""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def n_nodes(self) -> int:
        r"""The number of nodes (or vertices) in the graph"""
        raise NotImplementedError

    @property
    def n_edges(self) -> int:
        r"""The number of edges in the graph."""
        return len(self.edges())

    @abc.abstractmethod
    def adjacency_list(self) -> list[list]:
        r"""List containing the adjacency list of the graph where each node
        is represented by an integer in [0, n_nodes)"""
        raise NotImplementedError



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import permutations
from collections.abc import Sequence
import numpy as np

from .lattice import Lattice

from netket.utils.group import PointGroup, PGSymmetry, planar, cubic, Identity


def _perm_symm(perm: tuple) -> PGSymmetry:
    n = len(perm)
    M = np.zeros((n, n))
    M[range(n), perm] = 1
    return PGSymmetry(M)


def _axis_reflection(axis: int, ndim: int) -> PGSymmetry:
    M = np.eye(ndim)
    M[axis, axis] = -1
    return PGSymmetry(M)


def _grid_point_group(
    extent: Sequence[int], pbc: Sequence[bool], color_edges: bool
) -> PointGroup:
    """Point group of `Grid`, made up of axis permutations and flipping each axis."""
    ndim = len(extent)
    # Cannot exchange two axes if they are colored differently; otherwise,
    # can only exchange them if they have the same kind of BC and length.
    # Represent open BC by setting kind[i] = -extent[i], so just have to match these
    if color_edges:
        result = PointGroup([Identity()], ndim=ndim)
    else:
        axis_perm = []
        axes = np.arange(ndim, dtype=int)
        extent = np.asarray(extent, dtype=int)
        kind = np.where(pbc, extent, -extent)
        for perm in permutations(axes):
            if np.all(kind == kind[list(perm)]):
                if np.all(perm == axes):
                    axis_perm.append(Identity())
                else:
                    axis_perm.append(_perm_symm(perm))
        result = PointGroup(axis_perm, ndim=ndim)

    # reflections across axes and setting the origin
    # OBC axes are only symmetric w.r.t. their midpoint, (extent[i]-1)/2
    origin = []
    for i in range(ndim):
        result = result @ PointGroup([Identity(), _axis_reflection(i, ndim)], ndim=ndim)
        origin.append(0 if pbc[i] else (extent[i] - 1) / 2)

    return result.change_origin(origin)


def Grid(
    extent: Sequence[int],
    *,
    pbc: bool | Sequence[bool] = True,
    color_edges: bool = False,
    point_group: PointGroup | None = None,
    **kwargs,
) -> Lattice:
    """
    Constructs a hypercubic lattice given its extent in all dimensions.

    Args:
        extent: Size of the lattice along each dimension. It must be a list with
                integer components >= 1.
        pbc: If `True`, the grid will have periodic boundary conditions (PBC);
             if `False`, the grid will have open boundary conditions (OBC).
             This parameter can also be a list of booleans with same length as
             the parameter `length`, in which case each dimension will have
             PBC/OBC depending on the corresponding entry of `pbc`.
        color_edges: generates nearest-neighbour edges colored according to direction
                     i.e. edges along Cartesian direction #i have color i
                     cannot be used with `max_neighbor_order` or `custom_edges`
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the full hypercube symmetry group.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.

    Examples:
        Construct a 5x10 square lattice with periodic boundary conditions:

        >>> import netket
        >>> g=netket.graph.Grid(extent=[5, 10], pbc=True)
        >>> print(g.n_nodes)
        50

        Construct a 2x2x3 cubic lattice with open boundary conditions:

        >>> g=netket.graph.Grid(extent=[2,2,3], pbc=False)
        >>> print(g.n_nodes)
        12
    """
    extent = np.asarray(extent, dtype=int)

    ndim = len(extent)
    if isinstance(pbc, bool):
        pbc = [pbc] * ndim
    if color_edges:
        kwargs["custom_edges"] = [(0, 0, vec) for vec in np.eye(ndim)]
    if point_group is None:
        point_group = lambda: _grid_point_group(extent, pbc, color_edges)
    return Lattice(
        basis_vectors=np.eye(ndim),
        extent=extent,
        pbc=pbc,
        point_group=point_group,
        **kwargs,
    )


def Hypercube(length: int, n_dim: int = 1, *, pbc: bool = True, **kwargs) -> Lattice:
    r"""Constructs a hypercubic lattice with equal side length in all dimensions.
    Periodic boundary conditions can also be imposed.

    Args:
        length: Side length of the hypercube; must always be >=1
        n_dim: Dimension of the hypercube; must be at least 1.
        pbc: Whether the hypercube should have periodic boundary conditions
            (in all directions)
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.

    Examples:
         A 10x10x10 cubic lattice with periodic boundary conditions can be
         constructed as follows:

         >>> import netket
         >>> g = netket.graph.Hypercube(10, n_dim=3, pbc=True)
         >>> print(g.n_nodes)
         1000
    """
    if not isinstance(length, int) or length <= 0:
        raise TypeError("Argument `length` must be a positive integer")
    length_vector = [length] * n_dim
    return Grid(length_vector, pbc=pbc, **kwargs)


def Cube(length: int, *, pbc: bool = True, **kwargs) -> Lattice:
    """Constructs a cubic lattice of side `length`
    Periodic boundary conditions can also be imposed

    Args:
        length: Side length of the cube; must always be >=1
        pbc: Whether the cube should have periodic boundary conditions
            (in all directions)
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.

    Examples:
        A 10×10×10 cubic lattice with periodic boundary conditions can be
        constructed as follows:

        >>> import netket
        >>> g=netket.graph.Cube(10, pbc=True)
        >>> print(g.n_nodes)
        1000
    """
    return Hypercube(length, pbc=pbc, n_dim=3, **kwargs)


def Square(length: int, *, pbc: bool = True, **kwargs) -> Lattice:
    """Constructs a square lattice of side `length`
    Periodic boundary conditions can also be imposed

    Args:
        length: Side length of the square; must always be >=1
        pbc: Whether the square should have periodic boundary
            conditions (in both directions)
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.

    Examples:
        A 10x10 square lattice with periodic boundary conditions can be
        constructed as follows:

        >>> import netket
        >>> g=netket.graph.Square(10, pbc=True)
        >>> print(g.n_nodes)
        100
    """
    return Hypercube(length, pbc=pbc, n_dim=2, **kwargs)


def Chain(length: int, *, pbc: bool = True, **kwargs) -> Lattice:
    r"""Constructs a chain of `length` sites.
    Periodic boundary conditions can also be imposed

    Args:
        length: Length of the chain. It must always be >=1
        pbc: Whether the chain should have periodic boundary conditions
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.

    Examples:
        A 10 site chain with periodic boundary conditions can be
        constructed as follows:

        >>> import netket
        >>> g = netket.graph.Chain(10, pbc=True)
        >>> print(g.n_nodes)
        10
    """
    return Hypercube(length, pbc=pbc, n_dim=1, **kwargs)


def BCC(
    extent: Sequence[int],
    *,
    pbc: bool | Sequence[bool] = True,
    point_group: PointGroup | None = None,
    **kwargs,
) -> Lattice:
    """Constructs a BCC lattice of a given spatial extent.
    Periodic boundary conditions can also be imposed
    Sites are returned at the Bravais lattice points.

    Arguments:
        extent: Number of primitive unit cells along each direction, needs to be
            an array of length 3
        pbc: If `True`, the lattice will have periodic boundary conditions (PBC);
             if `False`, the lattice will have open boundary conditions (OBC).
             This parameter can also be a list of booleans with same length as
             the parameter `length`, in which case each dimension will have
             PBC/OBC depending on the corresponding entry of `pbc`.
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the full cubic symmetry group.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.


    Example:
        Construct a BCC lattice with 3×3×3 primitive unit cells:

        >>> from netket.graph import BCC
        >>> g = BCC(extent=[3,3,3])
        >>> print(g.n_nodes)
        27
    """
    basis = [[-0.5, 0.5, 0.5], [0.5, -0.5, 0.5], [0.5, 0.5, -0.5]]
    if point_group is None:
        # determine if full point group is realised by the simulation box
        point_group = cubic.Oh() if np.all(pbc) and len(set(extent)) == 1 else None
    return Lattice(
        basis_vectors=basis, extent=extent, pbc=pbc, point_group=point_group, **kwargs
    )


def FCC(
    extent: Sequence[int],
    *,
    pbc: bool | Sequence[bool] = True,
    point_group: PointGroup | None = None,
    **kwargs,
) -> Lattice:
    """Constructs an FCC lattice of a given spatial extent.
    Periodic boundary conditions can also be imposed
    Sites are returned at the Bravais lattice points.

    Arguments:
        extent: Number of primitive unit cells along each direction, needs
            to be an array of length 3
        pbc: If `True`, the lattice will have periodic boundary conditions (PBC);
             if `False`, the lattice will have open boundary conditions (OBC).
             This parameter can also be a list of booleans with same length as
             the parameter `length`, in which case each dimension will have
             PBC/OBC depending on the corresponding entry of `pbc`.
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the full cubic symmetry group.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.


    Example:
        Construct an FCC lattice with 3×3×3 primitive unit cells:

        >>> from netket.graph import FCC
        >>> g = FCC(extent=[3,3,3])
        >>> print(g.n_nodes)
        27
    """
    basis = [[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]]
    if point_group is None:
        # determine if full point group is realised by the simulation box
        point_group = cubic.Oh() if np.all(pbc) and len(set(extent)) == 1 else None
    return Lattice(
        basis_vectors=basis, extent=extent, pbc=pbc, point_group=point_group, **kwargs
    )


def Diamond(
    extent: Sequence[int],
    *,
    pbc: bool | Sequence[bool] = True,
    point_group: PointGroup | None = None,
    **kwargs,
) -> Lattice:
    """Constructs a diamond lattice of a given spatial extent.
    Periodic boundary conditions can also be imposed.

    Sites are returned at the 8a Wyckoff positions of the FCC lattice
    ([000], [1/4,1/4,1/4], and translations thereof).

    Arguments:
        extent: Number of primitive unit cells along each direction, needs to
            be an array of length 3
        pbc: If `True`, the lattice will have periodic boundary conditions (PBC);
             if `False`, the lattice will have open boundary conditions (OBC).
             This parameter can also be a list of booleans with same length as
             the parameter `length`, in which case each dimension will have
             PBC/OBC depending on the corresponding entry of `pbc`.
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the full symmetry group of the diamond lattice.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.


    Example:
        Construct a diamond lattice with 3×3×3 primitive unit cells:

        >>> from netket.graph import Diamond
        >>> g = Diamond(extent=[3,3,3])
        >>> print(g.n_nodes)
        54
    """
    basis = [[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]]
    sites = [[0, 0, 0], [0.25, 0.25, 0.25]]
    if point_group is None:
        # determine if full point group is realised by the simulation box
        point_group = cubic.Fd3m() if np.all(pbc) and len(set(extent)) == 1 else None
    return Lattice(
        basis_vectors=basis,
        site_offsets=sites,
        extent=extent,
        pbc=pbc,
        point_group=point_group,
        **kwargs,
    )


def Pyrochlore(
    extent: Sequence[int],
    *,
    pbc: bool | Sequence[bool] = True,
    point_group: PointGroup | None = None,
    **kwargs,
) -> Lattice:
    """Constructs a pyrochlore lattice of a given spatial extent.
    Periodic boundary conditions can also be imposed.

    Sites are returned at the 16c Wyckoff positions of the FCC lattice
    ([111]/8, [1 -1 -1]/8, [-1 1 -1]/8, [-1 -1 1]/8, and translations thereof).

    Arguments:
        extent: Number of primitive unit cells along each direction, needs to be
            an array of length 3
        pbc: If `True`, the lattice will have periodic boundary conditions (PBC);
            if `False`, the lattice will have open boundary conditions (OBC).
            This parameter can also be a list of booleans with same length as
            the parameter `length`, in which case each dimension will have
            PBC/OBC depending on the corresponding entry of `pbc`.
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the full symmetry group of the pyrochlore lattice.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.


    Example:
        Construct a pyrochlore lattice with 3×3×3 primitive unit cells:

        >>> from netket.graph import Pyrochlore
        >>> g = Pyrochlore(extent=[3,3,3])
        >>> print(g.n_nodes)
        108
    """
    basis = [[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]]
    sites = np.array([[1, 1, 1], [1, 3, 3], [3, 1, 3], [3, 3, 1]]) / 8
    if point_group is None:
        # determine if full point group is realised by the simulation box
        point_group = cubic.Fd3m() if np.all(pbc) and len(set(extent)) == 1 else None
    return Lattice(
        basis_vectors=basis,
        site_offsets=sites,
        extent=extent,
        pbc=pbc,
        point_group=point_group,
        **kwargs,
    )


def _hexagonal_general(
    extent,
    *,
    site_offsets=None,
    pbc: bool | Sequence[bool] = True,
    point_group: PointGroup | None = None,
    **kwargs,
) -> Lattice:
    basis = [[1, 0], [0.5, 0.75**0.5]]
    if point_group is None:
        # determine if full point group is realised by the simulation box
        point_group = planar.D(6) if np.all(pbc) and extent[0] == extent[1] else None
    return Lattice(
        basis_vectors=basis,
        extent=extent,
        site_offsets=site_offsets,
        pbc=pbc,
        point_group=point_group,
        **kwargs,
    )


def Triangular(extent, *, pbc: bool | Sequence[bool] = True, **kwargs) -> Lattice:
    r"""Constructs a triangular lattice of a given spatial extent.
    Periodic boundary conditions can also be imposed
    Sites are returned at the Bravais lattice points.

    Arguments:
        extent: Number of unit cells along each direction, needs to be an array
            of length 2
        pbc: If `True`, the lattice will have periodic boundary conditions (PBC);
             if `False`, the lattice will have open boundary conditions (OBC).
             This parameter can also be a list of booleans with same length as
             the parameter `length`, in which case each dimension will have
             PBC/OBC depending on the corresponding entry of `pbc`.
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the full hexagonal symmetry group.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.


    Example:
        Construct a triangular lattice with 3 × 3 unit cells:

        >>> from netket.graph import Triangular
        >>> g = Triangular(extent=[3, 3])
        >>> print(g.n_nodes)
        9
    """
    return _hexagonal_general(extent, site_offsets=None, pbc=pbc, **kwargs)


def Honeycomb(extent, *, pbc: bool | Sequence[bool] = True, **kwargs) -> Lattice:
    r"""Constructs a honeycomb lattice of a given spatial extent.
    Periodic boundary conditions can also be imposed.
    Sites are returned at the 2b Wyckoff positions.

    Arguments:
        extent: Number of unit cells along each direction, needs to be an array
            of length 2
        pbc: If `True`, the lattice will have periodic boundary conditions (PBC);
             if `False`, the lattice will have open boundary conditions (OBC).
             This parameter can also be a list of booleans with same length as
             the parameter `length`, in which case each dimension will have
             PBC/OBC depending on the corresponding entry of `pbc`.
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the full hexagonal symmetry group.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.


    Example:
        Construct a honeycomb lattice with 3 × 3 unit cells:

        >>> from netket.graph import Honeycomb
        >>> g = Honeycomb(extent=[3, 3])
        >>> print(g.n_nodes)
        18
    """
    return _hexagonal_general(
        extent,
        site_offsets=[[0.5, 0.5 / 3**0.5], [1, 1 / 3**0.5]],
        pbc=pbc,
        **kwargs,
    )


def Kagome(extent, *, pbc: bool | Sequence[bool] = True, **kwargs) -> Lattice:
    r"""Constructs a kagome lattice of a given spatial extent.
    Periodic boundary conditions can also be imposed.
    Sites are returned at the 3c Wyckoff positions.

    Arguments:
        extent: Number of unit cells along each direction, needs to be an array
            of length 2
        pbc: If `True`, the lattice will have periodic boundary conditions (PBC);
             if `False`, the lattice will have open boundary conditions (OBC).
             This parameter can also be a list of booleans with same length as
             the parameter `length`, in which case each dimension will have
             PBC/OBC depending on the corresponding entry of `pbc`.
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the full hexagonal symmetry group.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.


    Example:
        Construct a kagome lattice with 3 × 3 unit cells:

        >>> from netket.graph import Kagome
        >>> g = Kagome(extent=[3, 3])
        >>> print(g.n_nodes)
        27
    """
    return _hexagonal_general(
        extent,
        site_offsets=[[0.5, 0], [0.25, 0.75**0.5 / 2], [0.75, 0.75**0.5 / 2]],
        pbc=pbc,
        **kwargs,
    )


def KitaevHoneycomb(
    extent,
    *,
    pbc: bool | Sequence[bool] = True,
    point_group: PointGroup | None = None,
    **kwargs,
) -> Lattice:
    r"""Constructs a honeycomb lattice of a given spatial extent.

    Nearest-neighbour edges are coloured according to direction
    (cf. Kitaev, https://doi.org/10.1016/j.aop.2005.10.005).
    Periodic boundary conditions can also be imposed.
    Sites are returned at the 2b Wyckoff positions.

    Arguments:
        extent: Number of unit cells along each direction, needs to be an array
            of length 2
        pbc: If `True`, the lattice will have periodic boundary conditions (PBC);
             if `False`, the lattice will have open boundary conditions (OBC).
             This parameter can also be a list of booleans with same length as
             the parameter `length`, in which case each dimension will have
             PBC/OBC depending on the corresponding entry of `pbc`.
        point_group: point group object describing the symmetry of the lattice
            If not specified, uses the 180° rotation symmetry of the Kitaev model.
        kwargs: Additional keyword arguments are passed on to the constructor of
            :ref:`netket.graph.Lattice`.


    Example:
        Construct a Kitaev honeycomb lattice with 3 × 3 unit cells:

        >>> from netket.graph import KitaevHoneycomb
        >>> g = KitaevHoneycomb(extent=[3, 3])
        >>> print(g.n_nodes)
        18
        >>> print(len(g.edges(filter_color=2)))
        9
    """
    if point_group is None:
        point_group = planar.C(2) if np.all(pbc) else None
    return Lattice(
        basis_vectors=[[1, 0], [0.5, 0.75**0.5]],
        extent=extent,
        site_offsets=[[0.5, 0.5 / 3**0.5], [1, 1 / 3**0.5]],
        pbc=pbc,
        point_group=point_group,
        custom_edges=[
            (0, 1, [0.5, 0.5 / 3**0.5]),
            (0, 1, [-0.5, 0.5 / 3**0.5]),
            (0, 1, [0, -1 / 3**0.5]),
        ],
        **kwargs,
    )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import warnings
from textwrap import dedent


def bool_env(varname: str, default: bool) -> bool:
    """Read an environment variable and interpret it as a boolean.
    True: 'y', 'yes', 't', 'true', 'on', and '1';
    False: 'n', 'no', 'f', 'false', 'off', and '0'.

    Args:
        varname: the name of the variable
        default: the default boolean value
    """
    val = os.getenv(varname, str(default))
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif val in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError(f"invalid truth value {val!r} for environment {varname!r}")


def int_env(varname: str, default: int) -> int:
    """Read an environment variable and interpret it as an integer."""
    return int(os.getenv(varname, default))


def get_env(varname: str, type, default: int | bool) -> int | bool:
    if type is int:
        return int_env(varname, default)
    elif type is bool:
        return bool_env(varname, default)  # type: ignore
    else:
        raise TypeError(f"Unknown type {type}")


class ReadOnlyDict:
    def __init__(self, data):
        self._data = data

    def __getitem__(self, key):
        return self._data[key]


class Config:
    _HAS_DYNAMIC_ATTRIBUTES = True

    def __init__(self):
        self._values = {}
        self._types = {}
        self._editable_at_runtime = {}

        self._readonly = ReadOnlyDict(self._values)
        self._callbacks = {}

    def define(
        self,
        name,
        type,
        default,
        *,
        help,
        runtime=False,
        callback=None,
        lazy=False,
    ):  # noqa: W0613
        """
        Defines a new flag

        Args:
            name: the flag name, should be an uppercase string like "NETKET_XXX"
            type: should be the type (bool, int) of the flag
            default: default value
            help: a string to use as description of this flag
            runtime: whether the flag can be modified at runtime
            callback: an optional callback function taking the value as argument
                that is executed when the flag is changed
            lazy: do not call the callback at init.
        """
        if name in self._values:
            raise KeyError(f"Flag {name} already defined.")

        self._types[name] = type
        self._editable_at_runtime[name] = runtime
        self._values[name] = get_env(name, type, default)
        self._callbacks[name] = callback

        if callback is not None and not lazy:
            callback(self._values[name])

        @property
        def _read_config(self):
            return self.FLAGS[name]

        @_read_config.setter
        def _read_config(self, value):
            if self._callbacks[name] is not None:
                self._callbacks[name](value)
            self.update(name, value)

        setattr(Config, name.lower(), _read_config)

    @property
    def FLAGS(self):
        """
        The flags of this instance of netket
        """
        return self._readonly

    def update(self, name, value):
        """
        Updates a configuration variable in netket.

        Args:
            name: the name of the variable
            value: the new value
        """
        name = name.upper()

        if not self._editable_at_runtime[name]:
            raise RuntimeError(
                f"\n\nFlag `{name}` can only be set through an environment "
                "variable before importing netket.\n"
                "Try launching python with:\n\n"
                f"\t{name}={self.FLAGS[name]} python\n\n"
                "or execute the following snippet BEFORE importing netket:\n\n"
                "\t>>>import os\n"
                f'\t>>>os.environ["{name}"]="{self.FLAGS[name]}"\n'
                "\t>>>import netket as nk\n\n"
            )

        if not isinstance(value, self._types[name]):
            raise TypeError(
                f"Configuration {name} must be a {self._types[name]}, but the "
                f"value {value} is a {type(value)}."
            )

        self._values[name] = self._types[name](value)

    def __repr__(self):
        txt = "\nGlobal configurations for NetKet\n"
        for k, v in self._values.items():
            txt = txt + f" - {k} = {v}\n"

        return txt

    def __dir__(self):
        return list(k.lower() for k in self._values.keys())


config = Config()
FLAGS = config.FLAGS

config.define(
    "NETKET_DEBUG",
    bool,
    default=False,
    help="Enable debug logging in many netket functions.",
    runtime=True,
)

config.define(
    "NETKET_EXPERIMENTAL",
    bool,
    default=False,
    help="Enable experimental features.",
    runtime=True,
)

config.define(
    "NETKET_MPI_WARNING",
    bool,
    default=True,
    help=dedent(
        """
        Raise a warning when running python under MPI
        without mpi4py and other mpi dependencies installed.
        """
    ),
    runtime=False,
)

config.define(
    "NETKET_MPI",
    bool,
    default=True,
    help=dedent(
        """
        Prevent NetKet from using (and initializing) MPI. If this flag is
        `0` `mpi4py` and `mpi4jax` will not be imported.
        """
    ),
    runtime=False,
)

config.define(
    "NETKET_USE_PLAIN_RHAT",
    bool,
    default=False,
    help=dedent(
        """
        By default, NetKet uses the split-R̂ Gelman-Rubin diagnostic in `netket.stats.statistics`,
        which detects non-stationarity in the MCMC chains (in addition to the classes of
        chain-mixing failures detected by plain R) since version 3.4.
        Enabling this flag restores the previous behavior of using plain (non-split) Rhat.
        """
    ),
    runtime=True,
)

config.define(
    "NETKET_EXPERIMENTAL_FFT_AUTOCORRELATION",
    bool,
    default=False,
    help=dedent(
        """
        The integrated autocorrelation time $\tau_c$ is computed separately for each chain $c$.
        To summarize it for the user, `Stats.tau_corr` is changed to contain the average over all
        chains and a new field `Stats.tau_corr_max` is added containing the maximum autocorrelation
        among all chains (which helps to identify outliers). Using the average $\tau$ over all chains
        seems like a good choice as it results in a low-variance estimate
        (see [here](https://emcee.readthedocs.io/en/stable/tutorials/autocorr/#autocorr) for a good
        discussion).
        """
    ),
    runtime=True,
)

config.define(
    "NETKET_EXPERIMENTAL_DISABLE_ODE_JIT",
    bool,
    default=True,
    help=dedent(
        """
        Disables the jitting of the whole ode solver, mainly used within TDVP solvers.
        The jitting is sometimes incompatible with GPU-based calculations, and on large
        calculations it gives negligible speedups so it might be beneficial to disable it.
        """
    ),
    runtime=True,
)


def _setup_experimental_sharding_cpu(n_procs):
    if n_procs > 1:
        if "jax" in sys.modules:
            warnings.warn(
                "must load NetKet before jax if using experimental_sharding_cpu"
            )

        flags = os.environ.get("XLA_FLAGS", "")
        flags = f"{flags} --xla_force_host_platform_device_count={n_procs}"
        os.environ["XLA_FLAGS"] = flags


config.define(
    "NETKET_EXPERIMENTAL_SHARDING_CPU",
    int,
    default=0,
    help=dedent(
        """
        Set to >=1 to force JAX to use multiple threads as separate devices on cpu.
        Sets the XLA_FLAGS='--xla_force_host_platform_device_count=#' environment variable.
        Disabled by default.
        """
    ),
    runtime=False,
    callback=_setup_experimental_sharding_cpu,
)


def _update_x64(val):
    from jax import config as jax_config

    jax_config.update("jax_enable_x64", val)


# This flag is setup to mirror JAX_ENABLE_X64 with True default. any of the two
# Can be explicitly set.
config.define(
    "NETKET_ENABLE_X64",
    bool,
    default=bool_env(
        "JAX_ENABLE_X64", True
    ),  # respect explicit JAX_ENABLE_X64 settings
    help=dedent(
        """
        Enables double-precision for Jax. Equivalent to `JAX_ENABLE_X64` but defaults to
        True instead of False, as it is required throughout NetKet. By setting this flag
        to False NetKet will run without double-precision everywhere.
        """
    ),
    runtime=True,
    callback=_update_x64,
)


config.define(
    "NETKET_SPHINX_BUILD",
    bool,
    default=False,
    help=dedent(
        """
        Set to True when building documentation with Sphinx. Disables some decorators.
        """
    ),
    runtime=True,
)


def _setup_experimental_sharding(val):
    if val:
        from jax import config as jax_config

        jax_config.update("jax_threefry_partitionable", True)


config.define(
    "NETKET_EXPERIMENTAL_SHARDING",
    bool,
    default=int_env("NETKET_EXPERIMENTAL_SHARDING_CPU", 0) > 0,
    help=dedent(
        """
        Enables highly expermiental support of netket for running on multiple jax devices.

        Supports both multiple local devices, as well as global ones in a multi-process environment.
        See https://jax.readthedocs.io/en/latest/multi_process.html#initializing-the-cluster for
        how to initialize the latter.
        Distributes chains and samples equally among all available devices.

        Hybrid parallelization with MPI is not supported, enabling NETKET_EXPERIMENTAL_SHARDING
        disables mpi.
        """
    ),
    runtime=False,
    callback=_setup_experimental_sharding,
)


def _recompute_default_device(val):
    if val is False:
        import jax

        jax.config.update("jax_default_device", None)
    else:
        from netket.utils.mpi.gpu_autorank_util import autoset_default_gpu

        autoset_default_gpu()


config.define(
    "NETKET_MPI_AUTODETECT_LOCAL_GPU",
    bool,
    default=False,
    runtime=True,
    callback=_recompute_default_device,
    lazy=True,
    help=dedent(
        """
        If there are more than 1 device per MPI rank, and if they are GPU devices,
        Set the default device by querying the local MPI rank.
        """
    ),
)


config.define(
    "NETKET_RANDOM_STATE_FALLBACK_WARNING",
    bool,
    default=True,
    runtime=True,
    help=dedent(
        """
        Print a warning every time you use a fallback that could never stop running
        when using random_state of constrained hilbert spaces with custom
        constraints.
        """
    ),
)


config.define(
    "NETKET_EXPERIMENTAL_SHARDING_FAST_SERIALIZATION",
    bool,
    default=False,
    runtime=True,
    help=dedent(
        """
        If True (Defaults False) does not gather data on the master process when
        using flax.serialization methods. This allows to use orbax-checkpoint with
        higher efficiency.
        """
    ),
)


config.define(
    "NETKET_SPIN_ORDERING_WARNING",
    bool,
    default=True,
    runtime=True,
    help=dedent(
        """
        If True (Defaults True) warns if the ordering of spins in the Hilbert space
        is not declared.
        """
    ),
)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp

from netket.utils import struct
from netket.utils.types import Scalar


@struct.dataclass
class KahanSum(struct.Pytree):
    """
    Accumulator implementing Kahan summation [1], which reduces
    the effect of accumulated floating-point error.

    [1] https://en.wikipedia.org/wiki/Kahan_summation_algorithm
    """

    value: Scalar
    """
    Current value stored in this array
    """
    compensator: Scalar
    """
    Compensator used to fix addition/subtraction
    """

    def __init__(self, value, compensator=None) -> None:
        """
        Constructs the Kahan Summation accumulator with a zero-initialized
        compensator if unspecified.

        Args:
            value: The value to initialize this scalar.
            compensator: The state of the compensator. 0 by default.
        """
        self.value = jnp.asarray(value)
        if compensator is None:
            compensator = jnp.zeros_like(self.value)
        self.compensator = compensator

    def __add__(self, other: Scalar):
        delta = other - self.compensator
        new_value = self.value + delta
        new_compensator = (new_value - self.value) - delta
        return KahanSum(new_value, new_compensator)

    def __jax_array__(self, dtype=None):
        return self.value.astype(dtype)



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np

from jax import numpy as jnp

from netket.graph import AbstractGraph
from netket.hilbert import AbstractHilbert
from netket.jax import canonicalize_dtypes
from netket.utils.numbers import dtype as _dtype
from netket.utils.types import Array, DType

from .. import spin
from .._hamiltonian import SpecialHamiltonian
from .._local_operator import LocalOperator


class IsingBase(SpecialHamiltonian):
    r"""
    The Transverse-Field Ising Hamiltonian :math:`-h\sum_i \sigma_i^{(x)} +J\sum_{\langle i,j\rangle} \sigma_i^{(z)}\sigma_j^{(z)}`.

    This implementation is considerably faster than the Ising hamiltonian constructed by summing :class:`~netket.operator.LocalOperator` s.
    """

    def __init__(
        self,
        hilbert: AbstractHilbert,
        graph: AbstractGraph | Array,
        h: float,
        J: float,
        dtype: DType | None,
    ):
        r"""
        Constructs the Ising Operator from an hilbert space and a
        graph specifying the connectivity.

        Args:
            hilbert: Hilbert space the operator acts on.
            h: The strength of the transverse field.
            J: The strength of the coupling. Default is 1.0.
            dtype: The dtype of the matrix elements.

        Examples:
            Constructs an ``Ising`` operator for a 1D system.

            >>> import netket as nk
            >>> g = nk.graph.Hypercube(length=20, n_dim=1, pbc=True)
            >>> hi = nk.hilbert.Spin(s=0.5, N=g.n_nodes)
            >>> op = nk.operator.Ising(h=1.321, hilbert=hi, J=0.5, graph=g)
            >>> print(op)
            Ising(J=0.5, h=1.321; dim=20)
        """
        super().__init__(hilbert)

        dtype = canonicalize_dtypes(float, h, J, dtype=dtype)

        if isinstance(graph, AbstractGraph):
            if graph.n_nodes != hilbert.size:
                raise ValueError(
                    """
                    The size of the graph must match the hilbert space.
                    """
                )
            # support also a matrix input in here.
            graph = graph.edges()

        if isinstance(graph, list):
            graph = np.asarray(
                [[u, v] for u, v in graph],
                dtype=np.intp,
            )

        if graph.ndim != 2 or graph.shape[1] != 2:
            raise ValueError(
                """
                Graph should be one of:
                    - NetKet graph type (nk.operator.AbstractGraph)
                    - List of tuples, describing the edges
                    - a (N,2) array of integers.
                """
            )

        self._h = h.astype(dtype=dtype)
        self._J = J.astype(dtype=dtype)
        self._edges = graph.astype(np.intp)

    @property
    def h(self) -> float:
        """The magnitude of the transverse field"""
        return self._h

    @property
    def J(self) -> float:
        """The magnitude of the hopping"""
        return self._J

    @property
    def edges(self) -> np.ndarray:
        """The (N_conns, 2) matrix of edges on which the interaction term
        is non-zero.
        """
        return self._edges

    @property
    def is_hermitian(self) -> bool:
        """A boolean stating whether this hamiltonian is hermitian."""
        return True

    @property
    def dtype(self) -> DType:
        """The dtype of the matrix elements."""
        return jnp.promote_types(_dtype(self.h), _dtype(self.J))

    def conjugate(self, *, concrete=True):
        # if real
        if isinstance(self.h, float) and isinstance(self.J, float):
            return self
        else:
            raise NotImplementedError

    def n_conn(self, x, out=None):  # pragma: no cover
        r"""Return the number of states connected to x.

        Args:
            x (matrix): A matrix of shape (batch_size,hilbert.size) containing
                        the batch of quantum numbers x.
            out (array): If None an output array is allocated.

        Returns:
            array: The number of connected states x' for each x[i].

        """
        if out is None:
            out = np.empty(x.shape[0], dtype=np.int32)
        out.fill((self.h != 0) * x.shape[1] + 1)
        return out

    @property
    def max_conn_size(self) -> int:
        """The maximum number of non zero ⟨x|O|x'⟩ for every x."""
        return self.hilbert.size + 1

    def copy(self, *, dtype: DType | None = None):
        if dtype is None:
            dtype = self.dtype

        return type(self)(
            hilbert=self.hilbert, graph=self.edges, J=self.J, h=self.h, dtype=dtype
        )

    def to_local_operator(self):
        # The hamiltonian
        ha = LocalOperator(self.hilbert, dtype=self.dtype)

        if self.h != 0:
            for i in range(self.hilbert.size):
                ha -= self.h * spin.sigmax(self.hilbert, int(i), dtype=self.dtype)

        if self.J != 0:
            for i, j in self.edges:
                ha += self.J * (
                    spin.sigmaz(self.hilbert, int(i), dtype=self.dtype)
                    * spin.sigmaz(self.hilbert, int(j), dtype=self.dtype)
                )

        return ha

    def _iadd_same_hamiltonian(self, other):
        if self.hilbert != other.hilbert:
            raise NotImplementedError(
                "Cannot add hamiltonians on different hilbert spaces"
            )
        if not np.allclose(self.edges, other.edges):
            return NotImplemented

        self._h += other.h
        self._J += other.J
        return self

    def _isub_same_hamiltonian(self, other):
        if self.hilbert != other.hilbert:
            raise NotImplementedError(
                "Cannot add hamiltonians on different hilbert spaces"
            )

        if not np.allclose(self.edges, other.edges):
            return NotImplemented

        self._h -= other.h
        self._J -= other.J
        return self

    def __repr__(self):
        return (
            f"{type(self).__name__}(J={self._J}, h={self._h}; dim={self.hilbert.size})"
        )



# Copyright 2022 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import jax
import jax.numpy as jnp
from jax.scipy.special import logsumexp

from netket.utils.types import Array

from ._utils_dtype import dtype_complex


def logsumexp_cplx(a: Array, b: Array | None = None, **kwargs) -> jax.Array:
    """Compute the log of the sum of exponentials of input elements, always returning a
    complex number.

    Equivalent to, but more numerically stable than, `np.log(np.sum(b*np.exp(a)))`.
    If the optional argument `b` is omitted, `np.log(np.sum(np.exp(a)))` is returned.

    Wraps `jax.scipy.special.logsumexp` but uses `return_sign=True` if both `a` and `b`
    are real numbers in order to support `b<0` instead of returning `nan`.

    See the JAX function for details of the calling sequence;
    `return_sign` is not supported.
    """
    if jnp.iscomplexobj(a) or jnp.iscomplexobj(b):
        # logsumexp uses complex algebra anyway
        return logsumexp(a, b=b, **kwargs)
    else:
        a, sgn = logsumexp(a, b=b, **kwargs, return_sign=True)
        a = a + jnp.where(sgn < 0, 1j * jnp.pi, 0j)
        return a


def logdet_cmplx(A: Array) -> jax.Array:
    r"""Log-determinant, with automatic upconversion to a complex
    output dtype in order to encode the sign.

    This is a thin wrapper on top of {func}`jax.numpy.linalg.slogdet`.
    The mathematical formula is:

    .. math::

        \log(|A |)

    Args:
        A: A square matrix, or batch of matrices. The shape should be
            `(..., N, N)`

    Return:
        A scalar or batch of scalars with the smallest complex dtype
        computed from the input dtype. If the input has shape
        `(..., N, N)` the output has shape `(...)`

    """
    sign, logabsdet = jnp.linalg.slogdet(A)
    cplx_type = dtype_complex(A.dtype)
    return logabsdet.astype(cplx_type) + jnp.log(sign.astype(cplx_type))



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from netket.utils.dispatch import dispatch

from netket.operator import (
    AbstractSuperOperator,
    DiscreteOperator,
    Squared,
)

from netket.vqs.mc import kernels, get_local_kernel

from .state import MCMixedState


# Dispatches to select what expect-kernel to use
@dispatch
def get_local_kernel(  # noqa: F811
    vstate: MCMixedState, Ô: AbstractSuperOperator, chunk_size: int
):
    return kernels.local_value_kernel_chunked


@dispatch
def get_local_kernel(  # noqa: F811
    vstate: MCMixedState, Ô: Squared[AbstractSuperOperator], chunk_size: int
):
    return kernels.local_value_squared_kernel_chunked


@dispatch
def get_local_kernel(  # noqa: F811
    vstate: MCMixedState, Ô: DiscreteOperator, chunk_size: int
):
    return kernels.local_value_op_op_cost_chunked



# Copyright 2020, 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as _np
from scipy.sparse.linalg import bicgstab as _bicgstab
from scipy.sparse.linalg import LinearOperator as _LinearOperator

from .operator import AbstractOperator as _AbstractOperator
from jax.experimental.sparse import JAXSparse as _JAXSparse


def lanczos_ed(
    operator: _AbstractOperator,
    *,
    k: int = 1,
    compute_eigenvectors: bool = False,
    matrix_free: bool = False,
    scipy_args: dict | None = None,
):
    r"""Computes `first_n` smallest eigenvalues and, optionally, eigenvectors
    of a Hermitian operator using :meth:`scipy.sparse.linalg.eigsh`.

    Args:
        operator: NetKet operator to diagonalize.
        k: The number of eigenvalues to compute.
        compute_eigenvectors: Whether or not to return the
            eigenvectors of the operator. With ARPACK, not requiring the
            eigenvectors has almost no performance benefits.
        matrix_free: If true, matrix elements are computed on the fly.
            Otherwise, the operator is first converted to a sparse matrix.
        scipy_args: Additional keyword arguments passed to
            :meth:`scipy.sparse.linalg.eigvalsh`. See the Scipy documentation for further
            information.

    Returns:
        Either `w` or the tuple `(w, v)` depending on whether `compute_eigenvectors`
        is True.

        - w: Array containing the lowest `first_n` eigenvalues.
        - v: Array containing the eigenvectors as columns, such that`v[:, i]`
          corresponds to `w[i]`.

    Example:
        Test for 1D Ising chain with 8 sites.

        >>> import netket as nk
        >>> hi = nk.hilbert.Spin(s=1/2)**8
        >>> hamiltonian = nk.operator.Ising(hi, h=1.0, graph=nk.graph.Chain(8))
        >>> w = nk.exact.lanczos_ed(hamiltonian, k=3)
        >>> w
        array([-10.25166179, -10.05467898,  -8.69093921])
    """
    from scipy.sparse.linalg import eigsh

    actual_scipy_args = {}
    if scipy_args:
        actual_scipy_args.update(scipy_args)
    actual_scipy_args["which"] = "SA"
    actual_scipy_args["k"] = k
    actual_scipy_args["return_eigenvectors"] = compute_eigenvectors

    if matrix_free:
        # wrap the operator.to_linear_operator() in a scipy.sparse.linalg.LinearOperator
        n = operator.hilbert.n_states
        A = _LinearOperator(
            (n, n),
            operator.to_linear_operator().__matmul__,
            dtype=operator.dtype,
        )
    else:
        A = operator.to_sparse()
        if isinstance(A, _JAXSparse):
            # jax sparse arrays are not compatible with scipy eigsh.
            # wrap them in a scipy.sparse.linalg.LinearOperator
            A = _LinearOperator(A.shape, A.__matmul__, dtype=A.dtype)

    result = eigsh(A, **actual_scipy_args)

    if not compute_eigenvectors:
        # The sort order of eigenvalues returned by scipy changes based on
        # `return_eigenvalues`. Therefore we invert the order here so that the
        # smallest eigenvalue is still the first one.
        return result[::-1]
    else:
        return result


def full_ed(operator: _AbstractOperator, *, compute_eigenvectors: bool = False):
    """Computes all eigenvalues and, optionally, eigenvectors
    of a Hermitian operator by full diagonalization.

    Args:
        operator: NetKet operator to diagonalize.
        compute_eigenvectors: Whether or not to return the eigenvectors
            of the operator.

    Returns:
        Either `w` or the tuple `(w, v)` depending on whether `compute_eigenvectors`
        is True.

    Example:

        Test for 1D Ising chain with 8 sites.

        >>> import netket as nk
        >>> hi = nk.hilbert.Spin(s=1/2)**8
        >>> hamiltonian = nk.operator.Ising(hi, h=1.0, graph=nk.graph.Chain(8))
        >>> w = nk.exact.full_ed(hamiltonian)
        >>> w.shape
        (256,)
    """
    from numpy.linalg import eigh, eigvalsh

    dense_op = operator.to_dense()

    if compute_eigenvectors:
        return eigh(dense_op)
    else:
        return eigvalsh(dense_op)


def steady_state(lindblad, *, sparse=True, method="ed", rho0=None, **kwargs):
    r"""Computes the numerically exact steady-state of a lindblad master equation.
    The computation is performed either through the exact diagonalization of the
    hermitian :math:`L^\dagger L` matrix, or by means of an iterative solver (bicgstabl)
    targeting the solution of the non-hermitian system :math:`L\rho = 0`
    and :math:`\mathrm{Tr}[\rho] = 1`.

    Note that for systems with 7 or more sites it is usually computationally impossible
    to build the full lindblad operator and therefore only `iterative` will work.

    Note that for systems with hilbert spaces with dimensions above 40k, tol
    should be set to a lower value if the steady state has non-trivial correlations.

    Args:
        lindblad: The lindbladian encoding the master equation.
        sparse: Whether to use sparse matrices (default: False for ed, True for
            iterative)
        method: 'ed' (exact diagonalization) or 'iterative' (iterative bicgstabl)
        rho0: starting density matrix for the iterative diagonalization (default: None)
        kwargs...: additional kwargs passed to bicgstabl

    For full docs please consult SciPy documentation at
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.bicgstab.html

    Keyword Args:
        maxiter: maximum number of iterations for the iterative solver (default: None)
        tol: The precision for the calculation (default: 1e-05)
        callback: User-supplied function to call after each iteration. It is called as
            callback(xk), where xk is the current solution vector

    Returns:
        The steady-state density matrix.
    """
    M = lindblad.hilbert.physical.n_states

    if method == "ed":
        if not sparse:
            from numpy.linalg import eigh
            from warnings import warn

            warn(
                """For reasons unknown to me, using dense diagonalisation on this
                matrix results in very low precision of the resulting steady-state
                since the update to numpy 1.9.
                We suggest using sparse=True, however, if you wish not to, you have
                been warned.
                Your digits are your responsibility now.
                """,
                stacklevel=2,
            )

            lind_mat = lindblad.to_dense()

            ldagl = lind_mat.T.conj() * lind_mat
            w, v = eigh(ldagl)

        else:
            from scipy.sparse.linalg import eigsh

            lind_mat = lindblad.to_sparse()
            ldagl = lind_mat.T.conj() * lind_mat

            w, v = eigsh(ldagl, which="SM", k=2)

        print("Minimum eigenvalue is: ", w[0])
        rho = v[:, 0].reshape((M, M))
        rho = rho / rho.trace()

    elif method == "iterative":
        # An extra row is added at the bottom of the therefore M^2+1 long array,
        # with the trace of the density matrix. This is needed to enforce the
        # trace-1 condition.
        L = lindblad.to_linear_operator(sparse=sparse, append_trace=True)

        # Initial density matrix ( + trace condition)
        Lrho_start = _np.zeros((M**2 + 1), dtype=L.dtype)
        if rho0 is None:
            Lrho_start[0] = 1.0
            Lrho_start[-1] = 1.0
        else:
            Lrho_start[:-1] = rho0.reshape(-1)
            Lrho_start[-1] = rho0.trace()

        # Target residual (everything 0 and trace 1)
        Lrho_target = _np.zeros((M**2 + 1), dtype=L.dtype)
        Lrho_target[-1] = 1.0

        # Iterative solver
        print("Starting iterative solver...")
        res, info = _bicgstab(L, Lrho_target, x0=Lrho_start, **kwargs)

        rho = res[:-1].reshape((M, M))
        if info == 0:
            print("Converged trace is ", rho.trace())
        elif info > 0:
            print("Failed to converge after ", info, " ( trace is ", rho.trace(), " )")
        elif info < 0:
            print("An error occurred: ", info)

    else:
        raise ValueError("method must be 'ed' or 'iterative'")

    return rho



# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np

import jax
import jax.numpy as jnp

from netket.operator import DiscreteOperator
from netket.experimental.hilbert import SpinOrbitalFermions
from netket.utils.optional_deps import import_optional_dependency
from ._fermion_operator_2nd_numba import FermionOperator2nd


def compute_pyscf_integrals(mol, mo_coeff):
    t_ij_ao = jnp.asarray(mol.intor("int1e_kin") + mol.intor("int1e_nuc"))
    v_ijkl_ao = jnp.asarray(mol.intor("int2e").transpose(0, 2, 3, 1))
    c = jnp.asarray(mo_coeff.T)
    t_ij_mo = jax.jit(jnp.einsum, static_argnums=0)("ij,ai,bj->ab", t_ij_ao, c, c)
    v_ijkl_mo = jax.jit(jnp.einsum, static_argnums=0)(
        "ijkl,ai,bj,ck,dl->abcd", v_ijkl_ao, c, c, c, c
    )
    const = jnp.asarray(float(mol.energy_nuc()))
    return const, t_ij_mo, v_ijkl_mo


def spinorb_from_spatial_sparse_coo2(tij_sparse, interleave=False, spin_values=[0, 1]):
    sparse = import_optional_dependency("sparse", descr="TV_from_pyscf_molecule")

    # Σ_ijσ t_ij c†_iσ c_jσ
    # for σ ∈ spin_values
    # interleave=True -> 2i+spin
    # interleave=False -> n*spin+i

    if isinstance(spin_values, int):
        spin_values = list(spin_values)
    # assert spin_values in [[0], [1], [0,1]]

    assert isinstance(tij_sparse, sparse.COO)
    n = tij_sparse.shape[0]

    if interleave:
        t = lambda x: 2 * x
        tp1 = lambda x: 2 * x + 1
    else:
        t = lambda x: x
        tp1 = lambda x: x + n

    i, j = tij_sparse.coords
    a = tij_sparse.data
    a2 = []
    i2 = []
    j2 = []
    if 0 in spin_values:
        a2.append(a)
        i2.append(t(i))
        j2.append(t(j))
    if 1 in spin_values:
        a2.append(a)
        i2.append(tp1(i))
        j2.append(tp1(j))
    i2 = np.concatenate(i2)
    j2 = np.concatenate(j2)
    a2 = np.concatenate(a2)
    return sparse.COO(np.array([i2, j2]), a2, shape=(2 * n, 2 * n))


def spinorb_from_spatial_sparse_coo4(
    vijkl_sparse, interleave=False, _order_preserving=False
):
    sparse = import_optional_dependency("sparse", descr="TV_from_pyscf_molecule")

    # Σ_ijklμσ v_ijkl c†_iμ c†_jσ c_kμ c_lσ
    # interleave=True -> 2i+spin
    # interleave=False -> n*spin+i
    assert isinstance(vijkl_sparse, sparse.COO)
    n = vijkl_sparse.shape[0]

    if interleave:
        t = lambda x: 2 * x
        tp1 = lambda x: 2 * x + 1
    else:
        t = lambda x: x
        tp1 = lambda x: x + n

    p, q, r, s = vijkl_sparse.coords
    b = vijkl_sparse.data

    if _order_preserving:
        # swap i/j and k/l
        a2 = np.concatenate([-b, -b, b, b])
        p2 = np.concatenate([tp1(q), tp1(p), t(p), tp1(p)])
        q2 = np.concatenate([t(p), t(q), t(q), tp1(q)])
        r2 = np.concatenate([tp1(r), tp1(s), t(r), tp1(r)])
        s2 = np.concatenate([t(s), t(r), t(s), tp1(s)])
    else:
        # same as openfermion
        a2 = np.concatenate([b, b, b, b])
        p2 = np.concatenate([t(p), tp1(p), t(p), tp1(p)])
        q2 = np.concatenate([tp1(q), t(q), t(q), tp1(q)])
        r2 = np.concatenate([tp1(r), t(r), t(r), tp1(r)])
        s2 = np.concatenate([t(s), tp1(s), t(s), tp1(s)])

    return sparse.COO(
        np.array([p2, q2, r2, s2]), a2, shape=(2 * n, 2 * n, 2 * n, 2 * n)
    )


def to_desc_order_sparse(vijkl_sparse, cutoff, set_zero_same=True):
    sparse = import_optional_dependency("sparse", descr="TV_from_pyscf_molecule")

    # !! use this only after adding spin, spinorb_from_spatial_sparse will be wrong
    # because the (implicitly assumed) symmetries of the tensor are not conserved

    # assume (1,1,0,0) are already index order (daggers are left/ij)
    # swap i/j k/l so that i>j and k>l

    # now swap the larger one to the left, will cause lots of them to cancel
    i, j, k, l = vijkl_sparse.coords
    a = vijkl_sparse.data.copy()
    a[np.where(i < j)] *= -1
    a[np.where(k < l)] *= -1
    if set_zero_same:
        # set to zero all those where we try to create / destroy two on the same orbital
        a[np.where(i == j)] *= 0
        a[np.where(k == l)] *= 0
    new_coords = np.array(
        [np.maximum(i, j), np.minimum(i, j), np.maximum(k, l), np.minimum(k, l)]
    )
    # use coo to merge same indices
    vijkl_sparse2 = sparse.COO(new_coords, a, shape=vijkl_sparse.shape)
    # we might have some new almost zeros from the cancellations, make sure they are 0
    new_coords2 = vijkl_sparse2.coords
    a2 = vijkl_sparse2.data
    mask = np.abs(a2) > cutoff
    return sparse.COO(new_coords2[:, mask], a2[mask], shape=vijkl_sparse.shape)


def spinorb_from_spatial_sparse(tij_sparse, vijkl_sparse, interleave=False):
    return spinorb_from_spatial_sparse_coo2(
        tij_sparse, interleave
    ), spinorb_from_spatial_sparse_coo4(vijkl_sparse, interleave)


def arrays_to_terms(
    const, tij_sparse, vijkl_sparse, term_conj2=(1, 0), term_conj4=(1, 0, 1, 0)
):
    ij = np.array(
        [
            tij_sparse.coords[0],
            np.ones_like(tij_sparse.coords[0], dtype=int) * term_conj2[0],
            tij_sparse.coords[1],
            np.ones_like(tij_sparse.coords[1], dtype=int) * term_conj2[1],
        ]
    ).T.reshape(-1, 2, 2)

    ijkl = np.array(
        [
            vijkl_sparse.coords[0],
            np.ones_like(vijkl_sparse.coords[0], dtype=int) * term_conj4[0],
            vijkl_sparse.coords[1],
            np.ones_like(vijkl_sparse.coords[1], dtype=int) * term_conj4[1],
            vijkl_sparse.coords[2],
            np.ones_like(vijkl_sparse.coords[2], dtype=int) * term_conj4[2],
            vijkl_sparse.coords[3],
            np.ones_like(vijkl_sparse.coords[3], dtype=int) * term_conj4[3],
        ]
    ).T.reshape(-1, 4, 2)
    terms = ij.tolist() + ijkl.tolist()
    weights = tij_sparse.data.tolist() + vijkl_sparse.data.tolist()
    return terms, weights, const


def operator_from_arrays(
    const,
    tij_sparse,
    vijkl_sparse,
    n_electrons,
    hi=None,
    cls=FermionOperator2nd,
    term_conj2=(1, 0),
    term_conj4=(1, 0, 1, 0),
):
    """
    H = const + Σ tᵢⱼ cᵢ†cⱼ + Σ vᵢⱼₖₗ cᵢ†cⱼcₖ†cₗ
    use term_conj4=(1,1,0,0) if you want cᵢ†cⱼ†cₖcₗ instead.
    use a zero matrix/tensor of the correct shape for tij and/or vijkl if you only want to use one of them
    """
    if hi is None:
        if isinstance(n_electrons, tuple):
            assert len(n_electrons) == 2
            hi = SpinOrbitalFermions(
                n_orbitals=tij_sparse.shape[0] // 2,
                s=1 / 2,
                n_fermions_per_spin=n_electrons,
            )
        else:
            hi = SpinOrbitalFermions(
                n_orbitals=tij_sparse.shape[0], n_fermions_per_spin=n_electrons
            )
    terms, weights, constant = arrays_to_terms(
        const, tij_sparse, vijkl_sparse, term_conj2=term_conj2, term_conj4=term_conj4
    )
    return cls(hi, terms, weights, constant)


def TV_from_pyscf_molecule(
    molecule,  # type: pyscf.gto.mole.Mole  # noqa: F821
    mo_coeff: np.ndarray,
    *,
    cutoff: float = 1e-11,
) -> tuple[...]:
    r"""
    Computes the nuclear repulsion energy :math:`E_{nuc}`, and the T and
    V tensors encoding the 1-body and 2-body terms in the electronic
    hamiltonian of a pyscf molecule using the specified molecular orbitals.

    The tensors returned correspond to the following expressions:

    .. math::

        \hat{H} = E_{nuc} + \sum_{ij} T_{ij} \hat{c}^\dagger_i\hat{c}_j +
            \sum_{ijkl} V_{ijkl} \hat{c}^\dagger_i\hat{c}_\dagger_j\hat{c}_k\hat{c}_l

    The electronic spin degree of freedom is encoded following the *NetKet convention*
    where the first :math:`N_{\downarrow}` values of the indices :math:`i,j,k,l` represent
    the spin down electrons, and the following :math:`N_{\uparrow}` values represent the
    spin up.

    .. note::

        In the `netket.experimental.operator.pyscf` module you can find some utility
        functions to convert from normal ordering to other orderings, but
        those are all internals so if you need them do copy-paste them somewhere else.

    Example:
        Constructs the hamiltonian for a Li-H molecule, using the `sto-3g` basis
        and the Boys orbitals

        >>> from pyscf import gto, scf, lo
        >>> import netket as nk; import netket.experimental as nkx
        >>>
        >>> geometry = [('Li', (0., 0., -1.5109/2)), ('H', (0., 0., 1.5109/2))]
        >>> mol = gto.M(atom=geometry, basis='STO-3G')
        >>>
        >>> # compute the boys orbitals
        >>> mf = scf.RHF(mol).run()  # doctest:+ELLIPSIS
            converged SCF energy = -7.86338...
        >>> mo_coeff = lo.Boys(mol).kernel(mf.mo_coeff)
        >>> ha = nkx.operator.pyscf.TV_from_pyscf_molecule(mol, mo_coeff)

    Args:
        molecule: The pyscf :class:`~pyscf.gto.mole.Mole` object describing the
            Hamiltonian
        mo_coeff: The molecular orbital coefficients determining the
            linear combination of atomic orbitals to produce the
            molecular orbitals. If unspecified this defaults to
            the hartree fock orbitals computed using :class:`~pyscf.scf.HF`.
        cutoff: Ignores all matrix elements in the `V` and `T` matrix that have
            magnitude less than this value. Defaults to :math:`10^{-11}`

    Returns:
        :code:`E,T,V`: a scalar and two numpy arrays, the first with 2 dimensions
        and the latter with 4 dimensions.

    """
    sparse = import_optional_dependency("sparse", descr="TV_from_pyscf_molecule")

    # Compute the T and V matrices
    E_nuc, tij, vijkl = compute_pyscf_integrals(molecule, mo_coeff)
    tij_sparse = sparse.COO.from_numpy(tij * (np.abs(tij) >= cutoff))
    vijkl_sparse = sparse.COO.from_numpy(vijkl * (np.abs(vijkl) >= cutoff))
    tij_sparse_spin, vijkl_sparse_spin = spinorb_from_spatial_sparse(
        tij_sparse, vijkl_sparse
    )
    vijkl_sparse_spin = to_desc_order_sparse(vijkl_sparse_spin, cutoff)

    return E_nuc, tij_sparse_spin, vijkl_sparse_spin


def from_pyscf_molecule(
    molecule,  # type: pyscf.gto.mole.Mole  # noqa: F821
    mo_coeff: np.ndarray | None = None,
    *,
    cutoff: float = 1e-11,
    implementation: DiscreteOperator = FermionOperator2nd,
) -> DiscreteOperator:
    r"""
    Construct a netket operator encoding the electronic hamiltonian of a pyscf
    molecule in a chosen orbital basis.

    Example:
        Constructs the hamiltonian for a Li-H molecule, using the `sto-3g` basis
        and the default Hartree-Fock molecular orbitals.

        >>> from pyscf import gto, scf, fci
        >>> import netket as nk; import netket.experimental as nkx
        >>>
        >>> bond_length = 1.5109
        >>> geometry = [('Li', (0., 0., -bond_length/2)), ('H', (0., 0., bond_length/2))]
        >>> mol = gto.M(atom=geometry, basis='STO-3G')
        >>>
        >>> mf = scf.RHF(mol).run()  # doctest:+ELLIPSIS
            converged SCF energy = -7.86338...
        >>> E_hf = sum(mf.scf_summary.values())
        >>>
        >>> E_fci = fci.FCI(mf).kernel()[0]
        >>>
        >>> ha = nkx.operator.from_pyscf_molecule(mol)  # doctest:+ELLIPSIS
            converged SCF energy = -7.86338...
        >>> E0 = float(nk.exact.lanczos_ed(ha))
        >>> print(f"{E0 = :.5f}, {E_fci = :.5f}")
        E0 = -7.88253, E_fci = -7.88253

    Example:
        Constructs the hamiltonian for a Li-H molecule, using the `sto-3g` basis
        and the Boys orbitals using :class:`~pyscf.lo.Boys`.

        >>> from pyscf import gto, scf, lo
        >>> import netket as nk; import netket.experimental as nkx
        >>>
        >>> bond_length = 1.5109
        >>> geometry = [('Li', (0., 0., -bond_length/2)), ('H', (0., 0., bond_length/2))]
        >>> mol = gto.M(atom=geometry, basis='STO-3G')
        >>>
        >>> # compute the boys orbitals
        >>> mf = scf.RHF(mol).run()  # doctest:+ELLIPSIS
            converged SCF energy = -7.86338...
        >>> mo_coeff = lo.Boys(mol).kernel(mf.mo_coeff)
        >>> # use the boys orbitals to construct the netket hamiltonian
        >>> ha = nkx.operator.from_pyscf_molecule(mol, mo_coeff=mo_coeff)

    Args:
        molecule: The pyscf :class:`~pyscf.gto.mole.Mole` object describing the
            Hamiltonian
        mo_coeff: The molecular orbital coefficients determining the
            linear combination of atomic orbitals to produce the
            molecular orbitals. If unspecified this defaults to
            the hartree fock orbitals computed using :class:`~pyscf.scf.HF`.
        cutoff: Ignores all matrix elements in the `V` and `T` matrix that have
            magnitude less than this value. Defaults to :math:`10^{-11}`
        implementation: The particular implementation to use for the operator.
            Different fermionic operator implementation might have different
            performances. Defaults to
            :class:`netket.experimental.operator.FermionOperator2nd` (this might
            change in the future).

    Returns:
        A netket second quantised operator that encodes the electronic hamiltonian.
    """
    # If unspecified, use HF molecular orbitals
    if mo_coeff is None:
        pyscf = import_optional_dependency("pyscf", descr="pyscf_molecule_to_arrays")

        mf = pyscf.scf.HF(molecule).run()
        mo_coeff = mf.mo_coeff

    E_nuc, Tij, Vijkl = TV_from_pyscf_molecule(molecule, mo_coeff, cutoff=cutoff)

    ha = operator_from_arrays(
        E_nuc,
        Tij,
        0.5 * Vijkl,
        molecule.nelec,
        term_conj4=(1, 1, 0, 0),
        cls=implementation,
    )
    # TODO maybe run setup and set _max_conn_size here estimating it analytially
    return ha



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as jnp

from netket.utils.dispatch import parametric
from netket.utils.types import Array, DType

from .abstract_hilbert import AbstractHilbert
from .discrete_hilbert import DiscreteHilbert
from .index import is_indexable


@parametric
class DoubledHilbert(DiscreteHilbert):
    r"""
    Superoperatorial hilbert space for states living in the tensorised state
    :math:`\hat{H}\otimes \hat{H}`, encoded according to Choi's isomorphism.
    """

    def __init__(self, hilb: AbstractHilbert):
        r"""
        Superoperatorial hilbert space for states living in the tensorised
        state :math:`\hat{H}\otimes \hat{H}`, encoded according to Choi's
        isomorphism.

        Args:
            hilb: the Hilbert space H.

        Examples:
            Simple superoperatorial hilbert space for few spins.

            >>> import netket as nk
            >>> g = nk.graph.Hypercube(length=5,n_dim=2,pbc=True)
            >>> hi = nk.hilbert.Spin(N=3, s=0.5)
            >>> hi2 = nk.hilbert.DoubledHilbert(hi)
            >>> print(hi2.size)
            6
        """
        self.physical = hilb
        self._size = 2 * hilb.size

        super().__init__(shape=hilb.shape * 2)

    @property
    def size(self):
        return self._size

    @property
    def shape(self):
        return self._shape

    @property
    def is_finite(self):
        return self.physical.is_finite

    @property
    def is_indexable(self) -> bool:
        """Whether the space can be indexed with an integer"""
        n = self.physical.n_states
        return self.physical.is_indexable and is_indexable([n, n])

    @property
    def local_size(self):
        return self.physical.local_size

    @property
    def local_states(self):
        return self.physical.local_states

    @property
    def constrained(self):
        return self.physical.constrained

    def size_at_index(self, i):
        r"""Size of the local degrees of freedom for the i-th variable.

        Args:
            i: The index of the desired site

        Returns:
            The number of degrees of freedom at that site
        """
        return self.physical.size_at_index(
            i if i < self.physical.size else i - self.physical.size
        )

    def states_at_index(self, i):
        r"""A list of discrete local quantum numbers at the site i.
        If the local states are infinitely many, None is returned.

        Args:
            i: The index of the desired site.

        Returns:
            A list of values or None if there are infinitely many.
        """
        return self.physical.states_at_index(
            i if i < self.physical.size else i - self.physical.size
        )

    @property
    def size_physical(self):
        return self.physical.size

    @property
    def n_states(self):
        return self.physical.n_states**2

    def _numbers_to_states(self, numbers):
        # !!! WARNING
        # This code assumes that states are stored in a MSB
        # (Most Significant Bit) format.
        # We assume that the rightmost-half indexes the LSBs
        # and the leftmost-half indexes the MSBs
        # HilbertIndex-generated states respect this, as they are:
        # 0 -> [0,0,0,0]
        # 1 -> [0,0,0,1]
        # 2 -> [0,0,1,0]
        # etc...

        dim = self.physical.n_states
        left, right = jnp.divmod(numbers, dim)

        out_l = self.physical.numbers_to_states(left)
        out_r = self.physical.numbers_to_states(right)
        return jnp.concatenate([out_l, out_r], axis=-1)

    def _states_to_numbers(self, states):
        # !!! WARNING
        # See note above in numbers_to_states

        n = self.physical.size
        dim = self.physical.n_states

        _out_l = self.physical._states_to_numbers(states[:, 0:n])
        _out_r = self.physical._states_to_numbers(states[:, n : 2 * n])
        return _out_l * dim + _out_r

    def states_to_local_indices(self, x):
        return self.physical.states_to_local_indices(x)

    def local_indices_to_states(self, x: Array, dtype: DType = None):
        return self.physical.local_indices_to_states(x, dtype=dtype)

    def __repr__(self):
        return f"DoubledHilbert({self.physical})"

    @property
    def _attrs(self):
        return (self.physical,)



from typing import Protocol
import dataclasses
from functools import partial

import numpy as np

import warnings
import jax

from flax import serialization


# Specify the signature for the functions used to specify the deserialization logic
class ShardedDeserializationFunction(Protocol):
    def __call__(
        self, value_target: jax.Array, value_state: jax.Array, *, name: str = "."
    ) -> jax.Array: ...


@dataclasses.dataclass
class ShardedFieldSpec:
    """Specification of a sharded field.

    Used to specify (for the time being) how to handle serialization/deserialization (using
    flax.serialize) of a field in a NetKet-style Pytree.
    """

    sharded: bool = True
    """Boolean indicating whether the field is sharded or not. If False, all other
    fields are ignored.
    """

    mpi_sharded_axis: int | None = 0
    """
    Optional integer indicating which axis is to be considered sharded when running with
    MPI. Defaults to 0.

    If this is specified, loading with MPI a state saved from a run with jax sharding will
    scatter/partition the data along the specified axis.

    If None, MPI is not supported.
    """

    deserialization_function: ShardedDeserializationFunction | str | None = "relaxed"
    """
    Function to use to deserialize the data. Can be a callable with signature
    `def f(value_target: jax.Array, value_state: jax.Array, *, name: str = ".") -> jax.Array`
    or one of the following strings:
        - "fail": Raise an error if the sharded data does not match the target data.
        - "strict": Raise an error if the sharded data does not match the target data.
        - "relaxed": Ignore extra data in the sharded data if the target data is smaller
            than the serialised data; error if the target data is larger.
        - "relaxed-ignore-errors": Ignore extra data in the sharded data if the target data is
            smaller, and do nothing if the target data is larger.
        - "relaxed-rng-key": Special case for RNG keys, where we can safely truncate the
            serialized data if the target data is larger.

    The default is "relaxed".
    """

    def __post_init__(self):
        if isinstance(self.deserialization_function, type(None)):
            setattr(self, "deserialization_function", "fail")
        if isinstance(self.deserialization_function, str):
            if self.deserialization_function == "fail":
                setattr(
                    self, "deserialization_function", from_flax_state_dict_sharding_fail
                )
            elif self.deserialization_function == "strict":
                setattr(
                    self,
                    "deserialization_function",
                    from_flax_state_dict_sharding_strict,
                )
            elif self.deserialization_function == "relaxed":
                setattr(
                    self,
                    "deserialization_function",
                    from_flax_state_dict_sharding_relaxed,
                )
            elif self.deserialization_function == "relaxed-ignore-errors":
                setattr(
                    self,
                    "deserialization_function",
                    partial(from_flax_state_dict_sharding_relaxed, ignore_errors=True),
                )
            elif self.deserialization_function == "relaxed-rng-key":
                setattr(
                    self,
                    "deserialization_function",
                    from_flax_state_dict_sharding_relaxed_rng_key,
                )
            else:
                raise ValueError(
                    f"Unknown deserialization function logic {self.deserialization_function}."
                    "Valid values are 'fail/strict/relaxed' or a custom callable."
                )

    def __bool__(self):
        # Make this behave as a boolean, so we use it as sharding specification
        return self.sharded


def to_flax_state_dict_sharding(sharded_data):
    from netket.utils import mpi

    # This exists for backward compatibility...
    from netket.jax.sharding import gather
    from netket import config as nkconfig

    if mpi.n_nodes > 1:
        gathered_data = mpi.mpi_gather(sharded_data)
        result = gathered_data.reshape((-1,) + gathered_data.shape[2:])
    elif (
        nkconfig.netket_experimental_sharding
        and not nkconfig.netket_experimental_sharding_fast_serialization
    ):
        # 'old' behaviour (compatibility for now... disabled with the)
        # flag above
        result = gather(sharded_data)
    else:
        result = sharded_data

    return result


# Fail mode
def from_flax_state_dict_sharding_fail(value_target, value_state, *, name="."):
    raise RuntimeError(
        f"Deserialization for this sharded field [{name}] was disabled by "
        "the deserialization specification of the Pytree."
    )


# Strict mode
def from_flax_state_dict_sharding_strict(value_target, value_state, *, name="."):
    # sharding imports
    from netket.utils import mpi

    sharded_dim = 0

    if mpi.n_nodes > 1:
        # Assuming every rank got the full data... So we don't need to broadcast
        if (
            value_state.shape[sharded_dim]
            != mpi.n_nodes * value_target.shape[sharded_dim]
        ):
            raise ValueError(
                f"Sharded data ({name}) has shape {value_state.shape} but target has shape {value_target.shape}."
                "This probably means that you serialized the data with a different number of nodes or a different "
                "array size alltogether."
            )
        scattered_value_state = value_state.reshape(
            (mpi.n_nodes,) + value_target.shape
        )[mpi.rank]
        result = serialization.from_state_dict(
            value_target, scattered_value_state, name=name
        )
        result = jax.lax.with_sharding_constraint(result, value_target.sharding)

    else:  # if processes > 1
        unsharded_update = serialization.from_state_dict(
            value_target, value_state, name=name
        )
        result = jax.lax.with_sharding_constraint(
            unsharded_update, value_target.sharding
        )
    return result


def from_flax_state_dict_sharding_relaxed(
    value_target, value_state, *, name=".", ignore_errors: bool = False
):
    # sharding imports
    from netket.utils import mpi

    sharded_dim = 0

    serialized_total_sharded_size = (
        value_state.shape[sharded_dim] if value_state.ndim > 0 else 0
    )
    target_total_sharded_size = (
        mpi.n_nodes * value_target.shape[sharded_dim]
        if value_target.ndim > 0
        else mpi.n_nodes
    )

    if serialized_total_sharded_size > target_total_sharded_size:
        if mpi.rank == 0 and jax.process_index() == 0:
            warnings.warn(
                f"Sharded data ({name}) has shape {value_state.shape} but target has shape {value_target.shape}."
                "This probably means that you serialized the data with more nodes. We will try to load it anyway "
                "by ignoring the extra data."
            )
        if value_target.ndim == 0:
            value_state = value_state[: mpi.n_nodes]
        else:
            value_state = value_state[: mpi.n_nodes * value_target.shape[sharded_dim]]
    elif serialized_total_sharded_size < target_total_sharded_size:
        if ignore_errors:
            return value_target
        raise ValueError(
            f"\n"
            f"Sharded (or mpi distributed) data for field '{name}' size mismatch: \n"
            f"\t Info                                          \n"
            f"\t       sharded axis #           : {sharded_dim}\n"
            f"\t       # mpi nodes              : {mpi.n_nodes}\n"
            f"\t       # jax processes          : {jax.process_count()}\n"
            f"\t serialized shape   (global)    : {value_state.shape}\n"
            f"\t           (total_sharded_size) : {serialized_total_sharded_size}\n"
            f"\t restore trgt shape (global)    : {(mpi.n_nodes,) + value_target.shape}\n"
            f"\t                    (per-shard) : {value_target.shape}\n"
            f"\t           (total_sharded_size) : {target_total_sharded_size}\n"
            "This probably means that you serialized the data with less nodes. We cannot load the data in this case."
            "\n"
        )

    if mpi.n_nodes > 1:
        # Assuming every rank got the full data... So we don't need to broadcast
        scattered_value_state = value_state.reshape(
            (mpi.n_nodes,) + value_target.shape
        )[mpi.rank]
        result = serialization.from_state_dict(
            value_target, scattered_value_state, name=name
        )
        result = jax.lax.with_sharding_constraint(result, value_target.sharding)

    else:  # if processes > 1
        unsharded_update = serialization.from_state_dict(
            value_target, value_state, name=name
        )
        result = jax.lax.with_sharding_constraint(
            unsharded_update, value_target.sharding
        )
    return result


def from_flax_state_dict_sharding_relaxed_rng_key(
    value_target, value_state, *, name=".", ignore_errors: bool = False
):
    """
    Equivalent to `from_flax_state_dict_sharding_relaxed`, but special cased
    for rng keys, which we know that can be safely truncated if the target state
    is larger than the serialized state.
    """
    # sharding imports
    from netket.utils import mpi

    sharded_dim = 0

    serialized_total_sharded_size = (
        value_state.shape[sharded_dim] if value_state.ndim > 0 else 0
    )
    target_total_sharded_size = (
        mpi.n_nodes * value_target.shape[sharded_dim]
        if value_target.ndim > 0
        else mpi.n_nodes
    )

    if serialized_total_sharded_size > target_total_sharded_size:
        if mpi.rank == 0 and jax.process_index() == 0:
            warnings.warn(
                f"""
                Serialized RNG key data ({name}) has a BIGGER shape {value_state.shape} that the
                target to be restored to ({value_target.shape}).

                We will truncate the extra data.

                The loading will succeed, but the random numbers generated will be different
                from the original ones (this should usually not be a problem).
                """
            )
        if value_target.ndim == 0:
            value_state = value_state[: mpi.n_nodes]
        else:
            value_state = value_state[: mpi.n_nodes * value_target.shape[sharded_dim]]
    elif serialized_total_sharded_size < target_total_sharded_size:
        if mpi.rank == 0 and jax.process_index() == 0:
            warnings.warn(
                f"""
                Serialized RNG key data ({name}) has a SMALLER shape {value_state.shape} that the
                target to be restored to ({value_target.shape}).

                We will use the available data to construct a new, larger RNG key.

                The loading SHOULD succeed, but the random numbers generated will be different
                from the original ones (this should usually not be a problem).
                If it fails, open an issue.
                """
            )
        # construct the taget shape that we really-really want by multiplying the target
        # sharded dimension by number of mpi ranks if available
        target_shape = list(value_target.shape)
        target_shape[sharded_dim] = mpi.n_nodes * value_target.shape[sharded_dim]
        target_shape = tuple(target_shape)

        # Generate a new RNG key data with the target shape, using the current rng data as
        # a seed
        rng = np.random.default_rng(value_state)
        new_value_state = rng.integers(
            0,
            high=np.iinfo(value_state.dtype).max,
            dtype=value_state.dtype,
            size=target_shape,
        )
        # Copy the data from the serialized state to the new state
        new_value_state[: value_state.shape[0]] = value_state
        value_state = new_value_state

    if mpi.n_nodes > 1:
        # Assuming every rank got the full data... So we don't need to broadcast
        scattered_value_state = value_state.reshape(
            (mpi.n_nodes,) + value_target.shape
        )[mpi.rank]
        result = serialization.from_state_dict(
            value_target, scattered_value_state, name=name
        )
        result = jax.lax.with_sharding_constraint(result, value_target.sharding)

    else:  # if processes > 1
        unsharded_update = serialization.from_state_dict(
            value_target, value_state, name=name
        )
        result = jax.lax.with_sharding_constraint(
            unsharded_update, value_target.sharding
        )
    return result



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Ignore false-positives for redefined `product` functions:
# pylint: disable=function-redefined

import itertools

import numpy as np

from netket.utils import HashableArray, struct
from netket.utils.float import comparable, prune_zeros
from netket.utils.types import Array
from netket.utils.dispatch import dispatch

from ._semigroup import Element, FiniteSemiGroup, Identity


@struct.dataclass
class FiniteGroup(FiniteSemiGroup):
    """
    Collection of Elements expected to satisfy group axioms.
    Unlike FiniteSemiGroup, product tables, conjugacy classes, etc. can be calculated.

    Group elements can be implemented in any way, as long as a subclass of Group
    is able to implement their action. Subclasses must implement a :code:`_canonical()`
    method that returns an array of integers for each acceptable Element such that
    two Elements are considered equal iff the corresponding matrices are equal.

    """

    def __hash__(self):
        return super().__hash__()

    def _canonical(self, x: Element) -> Array:
        r"""
        Canonical form of :code:`Element`s, used for equality testing (i.e., two
        :code:`Element`s `x,y` are deemed equal iff
        :code:`_canonical(x) == _canonical(y)`.
        Must be overridden in subclasses

        Arguments:
            x: an `Element`

        Returns:
            the canonical form as a numpy.ndarray of integers
        """
        raise NotImplementedError

    def _canonical_array(self) -> Array:
        r"""
        Lists the canonical forms returned by `_canonical` as rows of a 2D array.
        """
        return np.array([self._canonical(x).flatten() for x in self.elems])

    def _canonical_lookup(self) -> dict:
        r"""
        Creates a lookup table from canonical forms to index in `self.elems`
        """
        return {
            HashableArray(self._canonical(element)): index
            for index, element in enumerate(self.elems)
        }

    def remove_duplicates(self, *, return_inverse=False) -> "FiniteGroup":
        r"""
        Returns a new :class:`FiniteGroup` with duplicate elements (that is,
        elements with identical canonical forms) removed.

        Arguments:
            return_inverse: If True, also return indices to reconstruct the
                original group from the result.

        Returns:
            The group with duplicate elements removed. If
            :code:`return_inverse==True` it also returns the list of indices
            needed to reconstruct the original group from the result.
        """
        result = np.unique(
            self._canonical_array(),
            axis=0,
            return_index=True,
            return_inverse=return_inverse,
        )
        group = FiniteGroup([self.elems[i] for i in sorted(result[1])])
        if return_inverse:
            return group, result[2]
        else:
            return group

    @struct.property_cached
    def inverse(self) -> Array:
        r"""
        Indices of the inverse of each element.

        Assuming the definitions

        .. code::

            g = self[idx_g]
            h = self[self.inverse[idx_g]]

        :code:`gh = product(g, h)` is equivalent to :code:`Identity()`
        """
        canonical_identity = self._canonical(Identity())
        inverse = np.zeros(len(self.elems), dtype=int)

        for i, e1 in enumerate(self.elems):
            for j, e2 in enumerate(self.elems):
                prod = e1 @ e2
                if np.all(self._canonical(prod) == canonical_identity):
                    inverse[i] = j

        return inverse

    @struct.property_cached
    def product_table(self) -> Array:
        r"""
        A table of indices corresponding to :math:`g^{-1} h` over the group.

        Assuming the definitions

        .. code::

            g = self[idx_g]
            h = self[idx_h]
            idx_u = self.product_table[idx_g, idx_h]

        :code:`self[idx_u]` corresponds to :math:`u = g^{-1} h` .
        """
        n_symm = len(self)
        product_table = np.zeros([n_symm, n_symm], dtype=int)

        lookup = self._canonical_lookup()

        for i, e1 in enumerate(self.elems[self.inverse]):
            for j, e2 in enumerate(self.elems):
                prod = e1 @ e2
                product_table[i, j] = lookup[HashableArray(self._canonical(prod))]

        return product_table

    @struct.property_cached
    def conjugacy_table(self) -> Array:
        r"""
        The conjugacy table of this Permutation Group.

        Assuming the definitions

        .. code::

            g = self[idx_g]
            h = self[idx_h]

        :code:`self[self.conjugacy_table[idx_g,idx_h]]`
        corresponds to :math:`h^{-1}gh`.
        """
        col_index = np.arange(len(self))[np.newaxis, :]
        # exploits that h^{-1}gh = (g^{-1} h)^{-1} h
        return self.product_table[self.product_table, col_index]

    @struct.property_cached
    def conjugacy_classes(self) -> tuple[Array, Array, Array]:
        r"""
        The conjugacy classes of the group.

        Returns:

            The three arrays

            - classes: a boolean array, each row indicating the elements that
              belong to one conjugacy class
            - representatives: the lowest-indexed member of each conjugacy class
            - inverse: the conjugacy class index of every group element

        """
        row_index = np.arange(len(self))[:, np.newaxis]

        # if is_conjugate[i,j] is True, self[i] and self[j] are in the same class
        is_conjugate = np.full((len(self), len(self)), False)
        is_conjugate[row_index, self.conjugacy_table] = True

        classes, representatives, inverse = np.unique(
            is_conjugate, axis=0, return_index=True, return_inverse=True
        )

        # Usually self[0] == Identity(), which is its own class
        # This class is listed last by the lexicographic order used by np.unique
        # so we reverse it to get a more conventional layout
        classes = classes[::-1]
        representatives = representatives[::-1]
        inverse = (representatives.size - 1) - inverse

        return classes, representatives, inverse

    @struct.property_cached
    def character_table_by_class(self) -> np.ndarray:
        r"""
        Calculates the character table using Burnside's algorithm.

        Each row of the output lists the characters of one irrep in the order the
        conjugacy classes are listed in `self.conjugacy_classes`.

        Assumes that :code:`Identity() == self[0]`, if not, the sign of some
        characters may be flipped. The irreps are sorted by dimension.
        """
        classes, _, _ = self.conjugacy_classes
        class_sizes = classes.sum(axis=1)
        # Construct a random linear combination of the class matrices c_S
        #    (c_S)_{RT} = #{r,s: r \in R, s \in S: rs = t}
        # for conjugacy classes R,S,T, and a fixed t \in T.
        #
        # From our oblique times table it is easier to calculate
        #    (d_S)_{RT} = #{r,t: r \in R, t \in T: rs = t}
        # for a fixed s \in S. This is just `product_table == s`, aggregated
        # over conjugacy classes. c_S and d_S are related by
        #    c_{RST} = |S| d_{RST} / |T|;
        # since we only want a random linear combination, we forget about the
        # constant |S| and only divide each column through with the appropriate |T|
        class_matrix = (
            classes @ random(len(self), seed=0)[self.product_table] @ classes.T
        )
        class_matrix /= class_sizes

        # The vectors |R|\chi(r) are (column) eigenvectors of all class matrices
        # the random linear combination ensures (with probability 1) that
        # none of them are degenerate
        _, table = np.linalg.eig(class_matrix)
        table = table.T / class_sizes

        # Normalise the eigenvectors by orthogonality: \sum_g |\chi(g)|^2 = |G|
        norm = np.sum(np.abs(table) ** 2 * class_sizes, axis=1, keepdims=True) ** 0.5
        table /= norm
        table /= _cplx_sign(table[:, 0])[:, np.newaxis]  # ensure correct sign
        table *= len(self) ** 0.5

        # Sort lexicographically, ascending by first column, descending by others
        sorting_table = np.column_stack((table.real, table.imag))
        sorting_table[:, 1:] *= -1
        sorting_table = comparable(sorting_table)
        _, indices = np.unique(sorting_table, axis=0, return_index=True)
        table = table[indices]

        # Get rid of annoying nearly-zero entries
        table = prune_zeros(table)

        return table

    def character_table(self) -> np.ndarray:
        r"""
        Calculates the character table using Burnside's algorithm.

        Each row of the output lists the characters of all group elements
        for one irrep, i.e. :code:`self.character_table()[i,g]` gives
        :math:`\chi_i(g)`.

        Assumes that :code:`Identity() == self[0]`, if not, the sign
        of some characters may be flipped. The irreps are sorted by dimension.
        """
        _, _, inverse = self.conjugacy_classes
        CT = self.character_table_by_class
        return CT[:, inverse]

    def character_table_readable(self) -> tuple[list[str], Array]:
        r"""
        Returns a conventional rendering of the character table.

        Returns:

            A tuple containing a list of strings and an array

            - :code:`classes`: a text description of a representative of
              each conjugacy class as a list
            - :code:`characters`: a matrix, each row of which lists the
              characters of one irrep
        """
        # TODO put more effort into nice rendering?
        classes, idx_repr, _ = self.conjugacy_classes
        class_sizes = classes.sum(axis=1)
        representatives = [
            f"{class_sizes[cls]}x{self[rep]}" for cls, rep in enumerate(idx_repr)
        ]
        return representatives, self.character_table_by_class

    @struct.property_cached
    def _irrep_matrices(self) -> list[Array]:
        # We use Dixon's algorithm (Math. Comp. 24 (1970), 707) to decompose
        # the regular representation of the group into its irreps.
        # We start with a Hermitian matrix E that commutes with every matrix in
        # this rep: the space spanned by its degenerate eigenvectors then all
        # transform according to some rep of the group.
        # The matrix is randomised to ensure there are no accidental degeneracies,
        # i.e. all these spaces are irreps.
        # For real irreps, real matrices are returned: if needed, the same
        # routine is run with a real symmetric and a complex Hermitian matrix.

        true_product_table = self.product_table[self.inverse]
        inverted_product_table = true_product_table[:, self.inverse]

        def invariant_subspaces(e, seed):
            # Construct a Hermitian matrix that commutes with all matrices
            # in the regular rep.
            # These matrices obey E_{g,h} = e_{gh^{-1}} for some vector e
            e = e[inverted_product_table]
            e = e + e.T.conj()

            # Since E commutes with all the ρ, its eigenspaces reduce the rep
            # For a random input vector, there are no accidental degeneracies
            # except for complex irreps and real symmetric E.
            e, v = np.linalg.eigh(e)

            # indices that split v into eigenspaces
            _, starting_idx = np.unique(comparable(e), return_index=True)

            # Calculate sᴴPv for one eigenvector v per eigenspace, a fixed
            # random vector s and all regular projectors P
            # These are calculated as linear combinations of sᴴρv for the
            # regular rep matrices ρ, which is given by the latter two terms
            vs = v[:, starting_idx]
            s = random(len(self), seed=seed)[inverted_product_table]
            # row #i of this `s` is sᴴρ(self[i]), where sᴴ is the random vector
            proj = self.character_table().conj() @ s @ vs
            starting_idx = list(starting_idx) + [len(self)]
            return v, starting_idx, proj

        # Calculate the eigenspaces for a real and/or complex random matrix
        # To check which are needed, calculate Frobenius-Schur indicators
        # This is +1, 0, -1 for real, complex, and quaternionic irreps
        squares = np.diag(true_product_table)
        frob = np.array(
            np.rint(
                np.sum(self.character_table()[:, squares], axis=1).real / len(self)
            ),
            dtype=int,
        )
        eigen = {}
        if np.any(frob == 1):
            # real irreps: start from a real symmetric invariant matrix
            e = random(len(self), seed=0)
            eigen["real"] = invariant_subspaces(e, seed=1)
        if np.any(frob != 1):
            # complex or quaternionic irreps: complex hermitian invariant matrix
            e = random(len(self), seed=2, cplx=True)
            eigen["cplx"] = invariant_subspaces(e, seed=3)

        irreps = []

        for i, chi in enumerate(self.character_table()):
            v, idx, proj = eigen["real"] if frob[i] == 1 else eigen["cplx"]
            # Check which eigenspaces belong to this irrep
            proj = np.logical_not(np.isclose(proj[i], 0.0))
            # Pick the first eigenspace in this irrep
            first = np.arange(len(idx) - 1, dtype=int)[proj][0]
            v = v[:, idx[first] : idx[first + 1]]
            # v is now the basis of a single irrep: project the regular rep onto it
            irreps.append(np.einsum("gi,ghj ->hij", v.conj(), v[true_product_table, :]))

        return irreps

    def irrep_matrices(self) -> list[Array]:
        """
        Returns matrices that realise all irreps of the group.

        Returns:
            A list of 3D arrays such that `self.irrep_matrices()[i][g]` contains
            the representation of `self[g]` consistent with the characters in
            `self.character_table()[i]`.
        """
        return self._irrep_matrices


def _cplx_sign(x):
    return x / np.abs(x)


def random(n, seed, cplx=False):
    if cplx:
        v = np.random.default_rng(seed).normal(size=(2, n))
        v = v[0] + 1j * v[1]
        return v
    else:
        return np.random.default_rng(seed).normal(size=n)


@dispatch
def product(A: FiniteGroup, B: FiniteGroup):
    return FiniteGroup(
        elems=[a @ b for a, b in itertools.product(A.elems, B.elems)],
    )



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time

import os
from os import path as _path

from flax import serialization

from netket.jax.sharding import extract_replicated

from .runtime_log import RuntimeLog


class JsonLog(RuntimeLog):
    """
      This logger serializes expectation values and other log data to a JSON file and can save the latest model parameters in MessagePack encoding to a separate file.

    It can be passed with keyword argument `out` to Monte Carlo drivers in order
    to serialize the output data of the simulation.

    This logger inherits from :class:`netket.logging.RuntimeLog`, so it maintains the dictionary
    of all logged quantities in memory, which can be accessed through the attribute
    :attr:`~netket.logging.JsonLog.data`.

    If the model state is serialized, then it can be de-serialized using the msgpack protocol
    of flax. For more information on how to de-serialize the output, see
    `here <https://flax.readthedocs.io/en/latest/flax.serialization.html>`_.
    The target of the serialization is the variational state itself.

    Data is serialized to json as several nested dictionaries. You can deserialize
    by simply calling :func:`json.load(open(filename)) <json.load>`.
    Logged expectation values will be captured inside histories objects, so they will
    have a subfield `iter` with the iterations at which that quantity has been computed,
    then `Mean` and others.
    Complex numbers are logged as dictionaries :code:`{'real':list, 'imag':list}`.
    """

    def __init__(
        self,
        output_prefix: str,
        mode: str = "write",
        save_params_every: int = 50,
        write_every: int = 50,
        save_params: bool = True,
        autoflush_cost: float = 0.005,
    ):
        """
        Construct a Json Logger.

        Args:
            output_prefix: the name of the output files before the extension
            save_params_every: every how many iterations should machine parameters be
                flushed to file
            write_every: every how many iterations should data be flushed to file
            mode: Specify the behaviour in case the file already exists at this
                output_prefix. Options are
                - `[w]rite`: (default) overwrites file if it already exists;
                - `[x]` or `fail`: fails if file already exists;
            save_params: bool flag indicating whether variables of the variational state
                should be serialized at some interval. The output file is overwritten
                every time variables are saved again.
            autoflush_cost: Maximum fraction of runtime that can be dedicated to
                serializing data. Defaults to 0.005 (0.5 per cent)
        """
        super().__init__()

        # Shorthands for mode
        if mode == "w":
            mode = "write"
        elif mode == "a":
            mode = "append"
        elif mode == "x":
            mode = "fail"

        if not ((mode == "write") or (mode == "append") or (mode == "fail")):
            raise ValueError(
                "Mode not recognized: should be one of `[w]rite`, `[a]ppend` or"
                "`[x]`(fail)."
            )

        if mode == "append":
            raise ValueError("Append mode is no longer supported.")

        file_exists = _path.exists(output_prefix + ".log") or _path.exists(
            output_prefix + ".mpack"
        )

        if file_exists and mode == "fail":
            raise ValueError(
                "Output file already exists. Either delete it manually or"
                "change `output_prefix`."
            )

        dir_name = _path.dirname(output_prefix)
        if dir_name != "":
            os.makedirs(dir_name, exist_ok=True)

        self._prefix = output_prefix
        self._file_mode = mode

        self._write_every = write_every
        self._save_params_every = save_params_every
        self._old_step = 0
        self._steps_notflushed_write = 0
        self._steps_notflushed_pars = 0
        self._save_params = save_params
        self._files_open = [output_prefix + ".log", output_prefix + ".mpack"]

        self._autoflush_cost = autoflush_cost
        self._last_flush_time = time.time()
        self._last_flush_runtime = 0.0
        self._last_flush_pars_time = time.time()
        self._last_flush_pars_runtime = 0.0

        self._flush_log_time = 0.0
        self._flush_pars_time = 0.0

    def __call__(self, step, item, variational_state=None):
        old_step = self._old_step
        super().__call__(step, item, variational_state)

        # Check if the time from the last flush is higher than the maximum
        # allowed runtime cost of flushing
        elapsed_time = time.time() - self._last_flush_time
        # On windows, the precision of `time.time` is much lower than that on Linux,
        # so `elapsed_time` may be essentially zero.
        # We add 1e-7 to avoid the zero division error.
        flush_anyway = (
            self._last_flush_runtime / (elapsed_time + 1e-7) < self._autoflush_cost
        )

        if (
            self._steps_notflushed_write % self._write_every == 0
            or step == old_step - 1
            or flush_anyway
        ):
            self._flush_log()

        elapsed_time = time.time() - self._last_flush_pars_time
        flush_anyway = (
            self._last_flush_pars_runtime / (elapsed_time + 1e-7) < self._autoflush_cost
        )

        if (
            self._steps_notflushed_pars % self._save_params_every == 0
            or step == old_step - 1
            or flush_anyway
        ):
            self._flush_params(variational_state)

        self._old_step = step
        self._steps_notflushed_write += 1
        self._steps_notflushed_pars += 1

    def _flush_log(self):
        # Time how long flushing data takes.
        self._last_flush_time = time.time()
        if self._is_master_process:
            self.serialize(self._prefix + ".log")
        self._last_flush_runtime = time.time() - self._last_flush_time

        self._flush_log_time += self._last_flush_runtime
        self._steps_notflushed_write = 0

    def _flush_params(self, variational_state):
        if not self._save_params:
            return
        if variational_state is None:
            return

        self._last_flush_pars_time = time.time()
        binary_data = serialization.to_bytes(
            extract_replicated(variational_state.variables)
        )
        if self._is_master_process:
            with open(self._prefix + ".mpack", "wb") as outfile:
                outfile.write(binary_data)
        self._last_flush_pars_runtime = time.time() - self._last_flush_pars_time

        self._flush_pars_time += self._last_flush_pars_runtime
        self._steps_notflushed_pars = 0

    def flush(self, variational_state=None):
        """
        Writes to file the content of this logger.

        Args:
            variational_state: optionally also writes the parameters of the machine.
        """
        self._flush_log()

        if variational_state is not None:
            self._flush_params(variational_state)

    def __del__(self):
        if hasattr(self, "_steps_notflushed_write"):
            if self._steps_notflushed_write > 0:
                self.flush()
        if hasattr(self, "_steps_notflushed_pars"):
            if self._steps_notflushed_pars > 0 and self._save_params:
                self.flush()

    def __repr__(self):
        _str = f"JsonLog('{self._prefix}', mode={self._file_mode}, "
        _str = _str + f"autoflush_cost={self._autoflush_cost})"
        _str = _str + "\n  Runtime cost:"
        _str = _str + f"\n  \tLog:    {self._flush_log_time}"
        _str = _str + f"\n  \tParams: {self._flush_pars_time}"
        return _str



# Copyright 2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import jax
import jax.numpy as jnp

from flax import struct
from flax import linen as nn

from netket import config
from netket.hilbert import TensorHilbert
from netket.utils.types import PyTree, PRNGKeyT

# Necessary for the type annotation to work
if config.netket_sphinx_build:
    from netket import sampler

from .base import MetropolisRule


class TensorRule(MetropolisRule):
    r"""A Metropolis sampling rule that can be used to combine different rules acting
    on different subspaces of the same tensor-hilbert space.
    """

    hilbert: TensorHilbert = struct.field(pytree_node=False)
    """The hilbert space upon which this rule is defined. This
    must be a :class:`nk.hilbert.TensorHilbert` with the same
    size as the expected input samples.
    """

    rules: tuple[MetropolisRule, ...]
    """Tuple of rules to be used on every partition of the hilbert
    space.
    """

    def __init__(
        self, hilbert: TensorHilbert, rules: tuple[MetropolisRule, ...]
    ) -> "TensorRule":
        r"""Construct the composition of rules.

        It should be constructed by passing a `TensorHilbert` space as a first argument
        and a list of rules as a second argument. Each `rule[i]` will be used to generate
        a transition for the `i`-th subspace of the tensor hilbert space.

        Args:
            hilbert: The tensor hilbert space on which the rule acts.
            rules: A list of rules, one for each subspace of the tensor hilbert space.
        """
        if not isinstance(hilbert, TensorHilbert):
            raise TypeError(
                "The Hilbert space of a `CombinedRule` must be a TensorHilbert,"
                "which is constructed as a product of different Hilbert spaces."
            )

        if not isinstance(rules, tuple | list) or not all(
            isinstance(r, MetropolisRule) for r in rules
        ):
            raise TypeError(
                "The second argument (rules) must be a tuple of `MetropolisRule` "
                f"rules, but you have passed {type(rules)}."
            )

        if len(hilbert.subspaces) != len(rules):
            raise ValueError(
                "Length mismatch between the rules and the hilbert space: Hilbert "
                f"has {len(hilbert.subspaces)} subpsaces, but you specified {len(rules)}."
            )

        self.hilbert = hilbert
        self.rules = tuple(rules)

    def init_state(
        self,
        sampler: "sampler.MetropolisSampler",  # noqa: F821
        machine: nn.Module,
        params: PyTree,
        key: PRNGKeyT,
    ) -> Any | None:
        N = self.hilbert._n_hilbert_spaces
        keys = jax.random.split(key, N)
        return tuple(
            self.rules[i].init_state(
                sampler.replace(hilbert=self.hilbert.subspaces[i]),
                machine,
                params,
                keys[i],
            )
            for i in range(N)
        )

    def reset(
        self,
        sampler: "sampler.MetropolisSampler",  # noqa: F821
        machine: nn.Module,
        params: PyTree,
        sampler_state: "sampler.SamplerState",  # noqa: F821
    ) -> Any | None:
        rule_states = []
        for i in range(self.hilbert._n_hilbert_spaces):
            # construct temporary sampler and rule state with correct sub-hilbert and
            # sampler-state objects.
            _sampler = sampler.replace(hilbert=self.hilbert.subspaces[i])
            _state = sampler_state.replace(rule_state=sampler_state.rule_state[i])

            rule_states.append(self.rules[i].reset(_sampler, machine, params, _state))
        return tuple(rule_states)

    def transition(self, sampler, machine, parameters, state, key, σ):
        keys = jax.random.split(key, self.hilbert._n_hilbert_spaces)

        σps = []
        log_prob_corr = []
        for i in range(self.hilbert._n_hilbert_spaces):
            σ_i = σ[..., self.hilbert._cum_indices[i] : self.hilbert._cum_sizes[i]]

            # construct temporary sampler and rule state with correct sub-hilbert and
            # sampler-state objects.
            _sampler = sampler.replace(hilbert=self.hilbert.subspaces[i])
            _state = state.replace(rule_state=state.rule_state[i])

            σps_i, log_prob_corr_i = self.rules[i].transition(
                _sampler, machine, parameters, _state, keys[i], σ_i
            )

            σps.append(σps_i)
            if log_prob_corr_i is not None:
                log_prob_corr.append(log_prob_corr_i)

        σp = jnp.concatenate(σps, axis=-1)
        log_prob_corr = sum(log_prob_corr) if len(log_prob_corr) > 0 else None
        return σp, log_prob_corr

    def __repr__(self):
        return "TensorRule(hilbert={self.hilbert}, rules={self.rules})"



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable


from netket.utils import StaticRange

from .homogeneous import HomogeneousHilbert


class CustomHilbert(HomogeneousHilbert):
    r"""A custom hilbert space with discrete local quantum numbers.

    .. warning::

        This class is deprecated in version 3.14. Use custom constraints with
        existing classes instead, or define your own hilbert space class.

        Do contact us on GitHub if you have any questions or need assistance.

    """

    def __init__(
        self,
        local_states: StaticRange | None,
        N: int = 1,
        constraint_fn: Callable | None = None,
    ):
        r"""
        Constructs a new ``CustomHilbert`` given a list of eigenvalues of the states and
        a number of sites, or modes, within this hilbert space.

        .. warning::

            This class is deprecated in version 3.14. Use custom constraints with
            existing classes instead, or define your own hilbert space class.

        Args:
            local_states: :class:`~netket.utils.StaticRange` object describing the
                numbers used to encode the local degree of freedom of this Hilbert
                Space.
            N: Number of modes in this hilbert space (default 1).
            constraint_fn: A function specifying constraints on the quantum numbers.
                Given a batch of quantum numbers it should return a vector
                of bools specifying whether those states are valid or not.

        The :class:`netket.utils.StaticRange` object works like a standard `range`
        object and is used to define the valid configurations of the local degrees
        of freedom.

        For example, the :class:`~netket.utils.StaticRange` of a Fock Hilbert space
        is constructed as

        .. code-block:: python

            >>> import netket as nk
            >>> n_max = 10
            >>> nk.utils.StaticRange(start=0, step=1, length=n_max)
            StaticRange(start=0, step=1, length=10, dtype=int64)

        and the range of a Spin-1/2 Hilbert space is constructed as:

        .. code-block:: python

            >>> import netket as nk
            >>> n_max = 10
            >>> nk.utils.StaticRange(start=-1, step=2, length=2)
            StaticRange(start=-1, step=2, length=2, dtype=int64)


        Examples:
            Simple custom hilbert space.

            >>> import netket as nk
            >>> g = nk.graph.Hypercube(length=10,n_dim=2,pbc=True)
            >>> local_states = nk.utils.StaticRange(start=-2.0, step=1.0, length=4)
            >>> hi = nk.hilbert.CustomHilbert(local_states=local_states, N=100)
            >>> print(hi.size)
            100
        """
        super().__init__(local_states, N, constraint_fn=constraint_fn)

    def _mul_sametype_(self, other):
        assert type(self) == type(other)
        if not self.constrained:
            if self.local_states == other.local_states:
                return CustomHilbert(self._local_states, self.size + other.size)

        return NotImplemented



import numpy as np

from netket.sampler.rules import ExchangeRule
from netket.graph import AbstractGraph
from netket.graph import disjoint_union
from netket.experimental.hilbert import SpinOrbitalFermions


class ParticleExchangeRule(ExchangeRule):
    """Exchange rule for particles on a lattice.

    Works similarly to :class:`netket.sampler.rules.ExchangeRule`, but
    takes into account that only occupied orbitals
    can be exchanged with unoccupied ones.

    This sampler conserves the number of particles.
    """

    def __init__(
        self,
        hilbert,
        *,
        clusters: list[tuple[int, int]] | None = None,
        graph: AbstractGraph | None = None,
        d_max: int = 1,
        exchange_spins: bool = False,
    ):
        r"""
        Constructs the ParticleExchange Rule.

        Particles are only exchanged between modes where the particle number is different.
        For fermions, only occupied orbitals can be exchanged with unoccupied ones.

        You can pass either a list of clusters or a netket graph object to
        determine the clusters to exchange.

        Args:
            hilbert: The hilbert space to be sampled.
            clusters: The list of clusters that can be exchanged. This should be
                a list of 2-tuples containing two integers. Every tuple is an edge,
                or cluster of sites to be exchanged.
            graph: A graph, from which the edges determine the clusters
                that can be exchanged.
            d_max: Only valid if a graph is passed in. The maximum distance
                between two sites
            exchange_spins: (default False) If exchange_spins, the graph must encode the
                connectivity  between the first N physical sites having same spin, and
                it is replicated using :func:`netket.graph.disjoint_union` other every
                spin subsector. This option conserves the number of fermions per
                spin subsector. If the graph does not have a number of sites equal
                to the number of orbitals in the hilbert space, this flag has no effect.
        """
        if not isinstance(hilbert, SpinOrbitalFermions):
            raise ValueError(
                "This sampler rule currently only works with SpinOrbitalFermions hilbert spaces."
            )
        if not exchange_spins and hilbert.n_spin_subsectors > 1:
            if graph is not None and graph.n_nodes == hilbert.n_orbitals:
                graph = disjoint_union(*[graph] * hilbert.n_spin_subsectors)
            if clusters is not None and np.max(clusters) < hilbert.n_orbitals:
                clusters = np.concatenate(
                    [
                        clusters + i * hilbert.n_orbitals
                        for i in range(hilbert.n_spin_subsectors)
                    ]
                )
        super().__init__(clusters=clusters, graph=graph, d_max=d_max)

    def __repr__(self):
        return f"ParticleExchangeRule(# of clusters: {len(self.clusters)})"



# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from netket.utils.types import DType

from netket.hilbert import AbstractHilbert

from ._abstract_observable import AbstractObservable


class AbstractOperator(AbstractObservable):
    """Abstract class for quantum Operators.

    An operator is a general object that defines a linear transformation
    on vectors of the Hilbert space and their expectation value can
    be comptued starting from a variational state using the method
    `expect` or `expect_and_grad` of the variational states.

    If the object is not a linear operator but some more general
    function, it should inherit from {class}`~netket.operator.AbstractObservable`
    instead.

    Operators over discrete hilbert spaces that can be converted
    to a dense representation should instead inherit from
    {class}`~netket.operator.DiscreteOperator`, while operators
    over that only work over continuous hilbert spaces should
    inherit from {class}`~netket.operator.ContinuousOperator`.

    This class determines the basic methods that an operator
    must implement to work correctly with NetKet.
    """

    def __init__(self, hilbert: AbstractHilbert):
        super().__init__(hilbert)

    @property
    def is_hermitian(self) -> bool:
        """Returns true if this operator is hermitian."""
        return False

    @property
    def H(self) -> "AbstractOperator":
        """Returns the Conjugate-Transposed operator"""
        if self.is_hermitian:
            return self

        from ._lazy import Adjoint

        return Adjoint(self)

    @property
    def T(self) -> "AbstractOperator":
        """Returns the transposed operator"""
        return self.transpose()

    @property
    @abc.abstractmethod
    def dtype(self) -> DType:
        """The dtype of the operator's matrix elements ⟨σ|Ô|σ'⟩."""

    def collect(self) -> "AbstractOperator":
        """
        Returns a guaranteed concrete instance of an operator.

        As some operations on operators return lazy wrappers (such as transpose,
        hermitian conjugate...), this is used to obtain a guaranteed non-lazy
        operator.
        """
        return self

    def transpose(self, *, concrete=False) -> "AbstractOperator":
        """Returns the transpose of this operator.

        Args:
            concrete: if True returns a concrete operator and not a lazy wrapper

        Returns:
            if concrete is not True, self or a lazy wrapper; the
            transposed operator otherwise
        """
        if not concrete:
            from ._lazy import Transpose

            return Transpose(self)
        else:
            raise NotImplementedError

    def conjugate(self, *, concrete=False) -> "AbstractOperator":
        """Returns the complex-conjugate of this operator.

        Args:
            concrete: if True returns a concrete operator and not a lazy wrapper

        Returns:
            if concrete is not True, self or a lazy wrapper; the
            complex-conjugated operator otherwise
        """
        raise NotImplementedError

    def conj(self, *, concrete=False) -> "AbstractOperator":
        return self.conjugate(concrete=False)

    def __repr__(self):
        return f"{type(self).__name__}(hilbert={self.hilbert}, dtype={self.dtype})"



# Copyright 2020, 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ._version import version as __version__  # noqa: F401

from .utils import config

from . import utils
from . import errors

__all__ = [
    "exact",
    "graph",
    "callbacks",
    "hilbert",
    "operator",
    "optimizer",
    "sampler",
    "stats",
    "utils",
    "vqs",
    "nn",
]


from . import jax
from . import stats

from . import graph
from . import hilbert

from . import nn

from . import (
    exact,
    callbacks,
    logging,
    operator,
    models,
    sampler,
    vqs,
    optimizer,
)


# Main applications
from .driver import VMC
from .driver import SteadyState



from flax.traverse_util import flatten_dict, unflatten_dict
from flax.core import unfreeze


# TODO(19-dec-2021): Deprecate: eventually remove this
def update_GCNN_parity(params):
    """Adds biases of parity-flip layers to the corresponding no-flip layers.
    Corrects for changes in GCNN_parity due to PR #1030 in NetKet 3.3.

    Args:
        params: a parameter pytree
    """
    # unfreeze just in case, doesn't break with a plain dict
    params = flatten_dict(unfreeze(params))
    to_remove = []
    for path in params:
        if (
            len(path) > 1
            and path[-2].startswith("equivariant_layers_flip")
            and path[-1] == "bias"
        ):
            alt_path = (
                *path[:-2],
                path[-2].replace("equivariant_layers_flip", "equivariant_layers"),
                path[-1],
            )
            params[alt_path] = params[alt_path] + params[path]
            to_remove.append(path)
    for path in to_remove:
        del params[path]
    return unflatten_dict(params)
